from aiohttp import web
from aiohttp_remotes import BasicAuth, setup
from jsonrpcserver.aio import methods
from concurrent.futures._base import CancelledError
try:
  from asyncio.base_futures import InvalidStateError
except ImportError:
  from asyncio.futures import InvalidStateError

import asyncio
import logging
import json
from uuid import uuid4
import base64
from secp256k1_zkp import PrivateKey
from os.path import split as path_split, join
from time import time

from jsonrpcserver.dispatcher import request_logger, response_logger
from aiohttp.web import access_logger

from leer import version
request_logger.setLevel(logging.ERROR)
response_logger.setLevel(logging.ERROR)
access_logger.setLevel(logging.ERROR)

class RPCManager():
  #Should be singleton?
  def __init__(self, loop, syncer, config):
    self.host = config['rpc']['host']
    self.port = config['rpc']['port']
    self.loop = loop
    self.syncer = syncer
    self.global_message_queue = self.syncer.queues['RPCManager']
    #self.allowed_clients
    self.requests = {}
    self.up = True
    self.logger = logging.getLogger("RPCManager")
    default_log_level = logging.INFO;
    if "logging" in config:#debug, info, warning, error, critical
      loglevels = { "debug":logging.DEBUG, "info":logging.INFO, "warning":logging.WARNING, "error":logging.ERROR, "critical":logging.CRITICAL}
      if "base" in config["logging"] and config["logging"]["base"] in loglevels:
        self.logger.setLevel(loglevels[config["logging"]["base"]])
      if "rpc" in config["logging"] and config["logging"]["rpc"] in loglevels:
        #its ok to rewrite
        self.logger.setLevel(loglevels[config["logging"]["rpc"]])

    rpc_manager_location = __file__
    web_wallet_dir = join(path_split(rpc_manager_location)[0], "web_wallet")
    self.app = web.Application(loop=self.loop)
    self.loop.run_until_complete(setup(self.app, BasicAuth(config['rpc']['login'],config['rpc']['password'], "realm", white_paths=['/'])))
    
    self.app.router.add_route('*', '/rpc', self.handle)
    self.app.router.add_route('*', '/', self.handle_root)
    self.app.router.add_static('/',web_wallet_dir)
    self.server = self.loop.create_server(self.app.make_handler(), self.host, self.port)
    asyncio.ensure_future(self.server, loop=loop)
    asyncio.ensure_future(self.check_queue())

    methods.add(self.ping)
    methods.add(self.getconnectioncount)
    methods.add(self.getheight)
    methods.add(self.getblocktemplate)
    methods.add(self.getwork)
    methods.add(self.submitwork)
    methods.add(self.validatesolution)
    methods.add(self.getbalancestats)
    methods.add(self.getbalancelist)
    methods.add(self.getbalance)
    methods.add(self.sendtoaddress)
    methods.add(self.getnewaddress)
    methods.add(self.dumpprivkey)
    methods.add(self.importprivkey)
    methods.add(self.getsyncstatus)
    methods.add(self.getblock)
    methods.add(self.getnodes)
    methods.add(self.connecttonode)
    methods.add(self.gettransactions)
    methods.add(self.getversion)


    methods.add(self.eth_getWork)
    methods.add(self.eth_submitWork)
    methods.add(self.eth_getBlockByNumber)
    self.no_auth_methods = ["eth_getWork", "eth_submitWork", "eth_getBlockByNumber"]

  async def handle(self, request):
    cors_origin_header = ("Access-Control-Allow-Origin", "*") #TODO should be restricted
    cors_headers_header = ("Access-Control-Allow-Headers", "content-type")
    if request.method=="OPTIONS":
      #preflight
      return web.Response(headers=[cors_origin_header, cors_headers_header])
    try:
      request = await request.text()
      response = await methods.dispatch(request, schema_validation=False)
    except CancelledError:
      return web.Response() #TODO can we set response.wanted to false?
    if response.is_notification:
        return web.Response()
    else:
        return web.json_response(response, status=response.http_status, headers=[cors_origin_header, cors_headers_header])

  async def handle_root(self, request):
    try:
      request_text = await request.text()
      if request_text=="":
        response = web.HTTPSeeOther("/index.html")
        return response
      else:
        try:
          rqst = json.loads(request_text)
          assert rqst["method"] in self.no_auth_methods
          response = await methods.dispatch(request_text, schema_validation=False)
          return web.json_response(response, status=response.http_status)
        except:
          return web.Response(request)
    except CancelledError:
      return web.Response() #TODO can we set response.wanted to false?    

  async def ping(self):
    return 'pong'

  async def getconnectioncount(self):
    _id = str(uuid4())
    self.syncer.queues['NetworkManager'].put({'action':'get connections num', 'id':_id, 'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']

  async def getheight(self):
    _id = str(uuid4())
    self.syncer.queues['Notifications'].put({'action':'get', 'id':_id, 'key': 'blockchain height','sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    if answer['result']=='error' or time() - answer['result']['time'] > 120:
      _id = str(uuid4())
      self.syncer.queues['Blockchain'].put({'action':'give tip height', 'id':_id, 'sender': "RPCManager"})
      self.requests[_id]=asyncio.Future()
      answer = await self.requests[_id]      
      self.requests.pop(_id)
      height = answer['result']
    else:
      height = answer['result']['value']
    return height

  async def getblocktemplate(self):
    _id = str(uuid4())
    self.syncer.queues['Blockchain'].put({'action':'give block template', 'id':_id, 'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return base64.b64encode(answer['result']).decode()

  async def getwork(self):
    _id = str(uuid4())
    self.syncer.queues['Blockchain'].put({'action':'give mining work', 'id':_id, 'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    if not answer["result"]=="error":
      res = ["0x"+answer["result"]["partial_hash"].upper(),\
             "0x"+answer["result"]["seed_hash"].upper(),\
             "0x"+(answer["result"]["target"]).upper(),\
             "0x"+answer["result"]["height"].to_bytes(8, "big").hex().upper()]
      return res
    else:
      return answer["error"]

  async def submitwork(self, hex_nonce, partial_hash_hex, compatibility_field):
    _id = str(uuid4())
    if "0x"==hex_nonce[:2]:
     hex_nonce = hex_nonce[2:]
    if "0x"==partial_hash_hex[:2]:
     partial_hash_hex = partial_hash_hex[2:]
    nonce = int(hex_nonce, 16).to_bytes(8, "big")
    partial_hash = bytes.fromhex(partial_hash_hex)
    self.syncer.queues['Blockchain'].put({'action':'take mining work', 'id':_id, 'sender': "RPCManager",
                                          'nonce':nonce, 'partial_hash':partial_hash})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    if answer["result"]=="error":
      return answer["error"]
    return answer['result']

  async def validatesolution(self, solution):
    _id = str(uuid4())
    self.syncer.queues['Blockchain'].put({'action':'take solved block template', 'id':_id,
                                          'sender': "RPCManager", 'solved template': base64.b64decode(solution.encode())})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']

  async def getbalancestats(self):
    _id = str(uuid4())
    self.syncer.queues['Wallet'].put({'action':'get confirmed balance stats', 'id':_id,
                                          'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']


  async def getbalancelist(self):
    _id = str(uuid4())
    self.syncer.queues['Wallet'].put({'action':'get confirmed balance list', 'id':_id,
                                          'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']

  async def getbalance(self):
    _id = str(uuid4())
    self.syncer.queues['Wallet'].put({'action':'get confirmed balance stats', 'id':_id,
                                          'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']['matured']['known_value']

  async def sendtoaddress(self, address, value):
    _id = str(uuid4())
    self.syncer.queues['Wallet'].put({'action':'generate tx', 'id':_id,
                                          'address': address, 'value': value,
                                          'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    if answer['result']=='error':
      return answer['error']
    _id = str(uuid4())
    self.syncer.queues['Blockchain'].put({'action':'add tx to mempool', 'id':_id,
                                          'tx': answer['result'],
                                          'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']


  async def getnewaddress(self):
    _id = str(uuid4())
    self.syncer.queues['Wallet'].put({'action':'give new taddress', 'id':_id,
                                          'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']

  async def dumpprivkey(self, address): #TODO DOESN'T WORK
    _id = str(uuid4())
    self.syncer.queues['Blockchain'].put({'action':'give private key', 'id':_id,
                                          'address':address, 'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']

  async def importprivkey(self, privkey): #TODO DOESN'T WORK
    _id = str(uuid4())
    self.syncer.queues['Blockchain'].put({'action':'take private key', 'id':_id,
                                          'privkey':privkey, 'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']

  async def getblock(self, block_num):
    _id = str(uuid4())
    self.syncer.queues['Blockchain'].put({'action':'give block info', 'id':_id,
                                          'block_num':block_num, 'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']

  async def getsyncstatus(self):
    _id = str(uuid4())
    self.syncer.queues['Notifications'].put({'action':'get', 'id':_id, 'key': 'blockchain height','sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    blockchain_height_answer = await self.requests[_id]
    self.requests.pop(_id)

    self.syncer.queues['Notifications'].put({'action':'get', 'id':_id, 'key': 'best header','sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    best_header_answer = await self.requests[_id]
    self.requests.pop(_id)

    self.syncer.queues['Notifications'].put({'action':'get', 'id':_id, 'key': 'best advertised height','sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    best_advertised_answer = await self.requests[_id]
    self.requests.pop(_id)
    
    self.syncer.queues['Notifications'].put({'action':'get', 'id':_id, 'key': 'core workload','sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    core_workload_answer = await self.requests[_id]
    self.requests.pop(_id)
    
    self.syncer.queues['Notifications'].put({'action':'get', 'id':_id, 'key': 'last wallet update','sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    last_wallet_update_answer = await self.requests[_id]
    self.requests.pop(_id)
    last_wallet_update = 0
    if not last_wallet_update_answer['result']=='error':
      last_wallet_update = last_wallet_update_answer['result']['value']
 
    response = {}

    if ('error' in [blockchain_height_answer['result'], best_header_answer['result'], best_advertised_answer['result']]) or \
       time() - min([blockchain_height_answer['result']['time'], best_header_answer['result']['time'], best_advertised_answer['result']['time']])>120:
      _id = str(uuid4())
      self.syncer.queues['Blockchain'].put({'action':'give synchronization status', 'id':_id,
                                          'sender': "RPCManager"})
      self.requests[_id]=asyncio.Future()
      answer = await self.requests[_id]
      self.requests.pop(_id)
      response = answer['result']
    else:
      response = {
                  'height': blockchain_height_answer['result']['value'],
                  'best_known_header': best_header_answer['result']['value'],
                  'best_advertised_header': best_advertised_answer['result']['value'],
                  'core_workload': core_workload_answer['result']['value']
                 }
    response['last_wallet_update'] = last_wallet_update;
    return response


  async def getnodes(self):
    _id = str(uuid4())
    self.syncer.queues['NetworkManager'].put({'action':'give my node', 'id':_id,
                                          'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    res = []
    if 'error' in answer['result']:
      return answer['result']
    for hp in answer['result']:
      host, port = hp
      static_key = base64.b64encode(answer['result'][hp]).decode()
      res.append({'host':str(host), 'port':port, 'static_key':static_key})

    _id = str(uuid4())
    self.syncer.queues['NetworkManager'].put({'action':'give nodes list', 'id':_id,
                                          'sender': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    if 'error' in answer['result']:
      return answer['result']
    for hp in answer['result']:
      host, port = hp
      static_key = base64.b64encode(answer['result'][hp]).decode()
      res.append({'host':str(host), 'port':str(port), 'static_key':static_key})
    return res

  async def connecttonode(self, node_str):
    _id = str(uuid4())
    (hp,sk)=node_str.split("@")
    host, port = hp.split(":")
    pub = base64.b64decode(sk.encode())
    self.syncer.queues['NetworkManager'].put({'action':'open connection', 'host':host,
         'port':port, 'static_key':pub, 'id':_id, 'request_source': "RPCManager"})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']

  async def gettransactions(self, num):
    _id = str(uuid4())
    self.syncer.queues['Wallet'].put({'action':'give last transactions info', 'id':_id,
                                      'sender': "RPCManager", 'num':int(num)})
    self.requests[_id]=asyncio.Future()
    answer = await self.requests[_id]
    self.requests.pop(_id)
    return answer['result']

  #Ethereum compatibility
  async def eth_getWork(self):
    _hash, seed, target, height = await self.getwork()
    # GPU miners work very bad with low diff target
    #dirty hack, to made minimal target 1m
    if not target[2:9]=="0000000":
      target = "0x0000003"+target[9:]
    return _hash, seed, target, height 

  async def eth_submitWork(self,hex_nonce, partial_hash_hex, mix_digest):
    res = await self.submitwork(hex_nonce, partial_hash_hex, mix_digest)
    return res=="Accepted"

  async def eth_getBlockByNumber(self, tag):
    '''
      Only returns block_number and difficulty in hex format.
    '''
    if not tag=='pending':
      return "Works only for pending tag"
    partial, seed, target, height = await self.getwork()
    difficulty = 2**256//int(target[2:], 16)
    return {'number':height.to_bytes(8, "big").hex(), 'difficulty':difficulty.to_bytes(8, "big").hex()}

  async def getversion(self):
    leer_version, leer_codename = version.VERSION, version.CODENAME
    return {'version':leer_version, 'codename':leer_codename}
    
 
  async def check_queue(self):
    while self.up:
      while not self.global_message_queue.empty():
        message = self.global_message_queue.get()
        if 'id' in message:
          if message['id'] in self.requests:
            try:
              self.requests[message['id']].set_result(message)
            except InvalidStateError:
              self.requests.pop(message['id'])
          elif message["action"] == "stop":
            self.loop.stop()    
        else:
          pass 
      await asyncio.sleep(0.2)
    

def RPCM_launcher(syncer, config):
  loop = asyncio.new_event_loop() #TODO use uvloop 
  asyncio.set_event_loop(loop)
  RPCM = RPCManager(loop, syncer, config)
  loop.run_forever()



from leer.core.utils import ObliviousDictionary

class VerificationCache:
  '''
   Verification cache is in memory cache for storing (non-context and semi-non-context) validity 
   of different objects.
   Two type of keys may be presented:
     tuple for semi-non-context checks. Second element of tuple is trated as block_number. If something
      was valid at height h, it should be valid at any height h' such as h'>=h. If something was not valid
      at height h, it should not be valid at any height h' such as h'<=h.
     anything else, in this case verification cache is just oblivious dictionary
  '''
  def __init__(self):
    self.od = ObliviousDictionary(3600)

  def __getitem__(self, _index):
    if isinstance(_index, tuple):
      _index, check_block = _index
      is_valid, has_block = self.od[_index]
      if is_valid and check_block>=has_block: #Once valid, always valid
        return True
      if (not is_valid) and (check_block<=has_block): #If not valid at block x, not valid for all blocks before x
        return False
      raise KeyError #Generally cant say anything
    else:
      return self.od[_index]

  def __setitem__(self, _index, _object):
    if isinstance(_index, tuple):
      _index, check_block = _index
      self.od[_index]=(_object, check_block)
    else:
      self.od[_index]=_object


verification_cache = VerificationCache()



import logging
import asyncio
import random

from concurrent.futures import CancelledError

from .network_node import NetworkNode
from .messages import message_id, inv_message_id

class Node(NetworkNode):
  def __init__(self, our_node, params, loop, message_handler, serialized_params=None):
    NetworkNode.__init__(self, our_node, params, loop, serialized_params=serialized_params);
    self.message_handler = message_handler
    self.ping_loop_task = None
    self.peer_exchange_loop_task = None

  async def on_established_connection(self):
    self.logger.info("Connection is established")
    print("Connection with %s:%s is established"%(self.host, self.port))
    init_send = await self.send(inv_message_id["init"] + self.our_node.serialize_params())
    if not init_send:
      self._on_closed_connection(error="Cant send init message")
      return
    await asyncio.sleep(0.2)
    give_nodes_send = await self.send(inv_message_id["give nodes"])
    if not give_nodes_send:
      self._on_closed_connection(error="Cant send give nodes message")
      return
    self.ping_loop_task = asyncio.ensure_future(self.ping_loop())
    self.peer_exchange_loop_task = asyncio.ensure_future(self.peer_exchange_loop())

  async def on_closed_connection(self):
    if self.ping_loop_task:
      self.ping_loop_task.cancel()
    if self.peer_exchange_loop_task:
      self.peer_exchange_loop_task.cancel()
    await self.message_handler(self,'close', b'\x00')

  async def ping_loop(self):
    try:
      if self.connected:
        await self.send(inv_message_id["ping"])
        await asyncio.sleep(random.randint(5, 10)) #TODO should check pong here?
        asyncio.ensure_future(self.ping_loop())
    except CancelledError:
      pass


  async def peer_exchange_loop(self):
    try:
      if self.connected:
        await self.send(inv_message_id["give nodes"])
        await asyncio.sleep(30) #TODO should check pong here?
        asyncio.ensure_future(self.peer_exchange_loop())
    except CancelledError:
      pass

  async def handle_message(self,message):
    if len(message)<2:
      pass #TODO DoS protection
    _id,message = bytes(message[:2]), message[2:]
    if not _id in message_id:
      pass #TODO DoS protection
    _type = message_id[_id]
    if _type == "ping":
      await self.send(inv_message_id["pong"])
    if _type == "pong":
      pass
    else:
      await self.message_handler(self,_type, message)

  def __repr__(self):
    return "<Node %s:%s  adv %s:%s>"%(self.host,self.port, self.advertised_host, self.advertised_port)



from leer.core.storage.merkle_storage import MMR
from leer.core.storage.key_value_storage import KeyValueStorage
from secp256k1_zkp import PedersenCommitment, PublicKey
from leer.core.lubbadubdub.ioput import IOput
from leer.core.utils import sha256
import os




class CommitmentMMR(MMR):
      def sum(self, x1,x2):
        # each index is 33 bytes for commitments and 32 for hash
        comm1, hash1 = x1[:33], x1[33:65]
        comm2, hash2 = x2[:33], x2[33:65]
        comm1, comm2 = PedersenCommitment(commitment=comm1, raw=True), PedersenCommitment(commitment=comm2, raw=True)
        #XXX we definetely need sum of pedersen commitments on libsecp256 level.
        pk= PublicKey()
        pk.combine([comm1.to_public_key().public_key, comm2.to_public_key().public_key])
        sm = pk.to_pedersen_commitment()
        first_part = sm.serialize()
        second_part = sha256(hash1+hash2)
        return first_part+second_part

class TXOMMR(MMR):
      def sum(self, x1,x2):
        return sha256(x1+x2)

class ConfirmedTXOStorage:
    '''
      Storage for all TXOs which are already in blocks.
      It is double MMR tree: first for commitments only, second for full txout
      Commitment tree:
        Each commitment leaf obj is `b""` and commitment leaf index is `bytes(serialized_apc)+bytes(hash(serialized_apc))`.
        Summ of two nodes in commitment tree = bytes(serialized_apc1+serialized_apc1)+hash(hash(serialized_apc)+hash(serialized_apc)),
        where commitments are summed as points on curve and summ of hashes is concantenation.

      Txout tree:
        Each txout leaf obj is serialized ioput, leaf index is hash(ioput_index).
        Summ of two nodes is hash of concantenation

    '''

    def __init__(self, path, env, wtx):
      self.commitments = CommitmentMMR("commitments", path, clear_only=False, env=env, wtx=wtx)
      self.txos = TXOMMR("txos", path, discard_only=True, env=env, wtx=wtx)
      self.burden = KeyValueStorage(name="burden", env=env, wtx=wtx)

    def get(self, hash_and_pc, rtx):
      res = self.txos.get_by_hash(sha256(hash_and_pc), rtx=rtx)
      if not res:
        raise KeyError(hash_and_pc)
      utxo=IOput()
      utxo.deserialize_with_context(res)
      utxo.set_verified_and_correct() #Trust saved outputs
      return utxo

    #TODO Full method removement is postponed till comments inside will be resolved
    #def __setitem__(self, hash_and_pc, utxo):
    #  #TODO __setitem__ interface should be substituted with append-like interface
    #  #XXX problem here: commitments indexes should be apc+hash(apc), not hash_and_pc
    #  #here we should save
    #  self.txos.append(_hash(hash_and_pc),utxo.serialize())
    #  self.commitments.append(utxo.commitment_index,b"")

    def append(self, utxo, wtx):
      assert utxo.verify() #Should be fast since cached
      assert utxo.address_excess_num_index
      self.txos.append(wtx=wtx, obj_index=sha256(utxo.serialized_index), obj=utxo.serialize_with_context())
      self.commitments.append_unique(wtx=wtx, obj_index=utxo.commitment_index, obj=b"")

    def spend(self, utxo, wtx, return_revert_obj=False):
      txos = self.txos.discard(sha256(utxo.serialized_index), wtx=wtx)
      commitment = self.commitments.clear(utxo.commitment_index, wtx=wtx)
      if return_revert_obj:
        return (txos, commitment)

    def find(self, hash_and_pc, rtx):
      '''
        In contrast with __getitem__ find will try to find even spent
        outputs for other (syncing) nodes.
      '''
      res = self.txos.find_by_hash(sha256(hash_and_pc), rtx=rtx)
      if not res:
        raise KeyError(hash_and_pc)
      utxo=IOput()
      utxo.deserialize_with_context(res)
      utxo.set_verified_and_correct() #Trust saved outputs
      return utxo

    def unspend(self, revert_obj, wtx):
      (txos, commitment) = revert_obj
      self.txos.revert_discarding(txos, wtx=wtx)
      self.commitments.revert_clearing(commitment, wtx=wtx)

    def has(self, serialized_index, rtx):
      return bool(self.txos.get_by_hash(sha256(serialized_index), rtx=rtx))

    def remove(self,n, wtx):
      self.commitments.remove(n,wtx=wtx)
      ser_removed_outputs = self.txos.remove(n,wtx=wtx)
      removed_outputs=[]
      for _ser in ser_removed_outputs:
        utxo=IOput()
        utxo.deserialize_with_context(_ser)
        removed_outputs.append(utxo)
      return removed_outputs

    def get_commitment_root(self, rtx):
      return self.commitments.get_root(rtx=rtx)

    def get_txo_root(self, rtx):
      return self.txos.get_root(rtx=rtx)

    def get_state(self, rtx):
      return self.commitments.get_state(rtx=rtx)

    def set_state(self, state, wtx):
      self.txos.set_state(state, wtx=wtx)
      self.commitments.set_state(state, wtx=wtx)

    def find_wo_deser(self, hash_and_pc, rtx):
      res = self.txos.find_by_hash(sha256(hash_and_pc), rtx=rtx)
      if not res:
        raise KeyError(hash_and_pc)
      return res


class TXOsStorage:

  class Interface:

    def __init__(self):
      self.storage = {}

    def __getitem__(self, hash_and_pc):
      if not hash_and_pc in self.storage:
        raise KeyError(hash_and_pc)
      return self.storage[hash_and_pc]

    def __setitem__(self, hash_and_pc, utxo):
      #TODO __setitem__ interface should be substituted with append-like interface
      #here we should save
      self.storage[hash_and_pc]=utxo

    def remove(self, utxo):
      self.remove_by_index(utxo.serialized_index)

    def remove_by_index(self, _index):
      self.storage.pop(_index)

    def __contains__(self, utxo):
      return utxo in self.storage  

    def flush(self):
      self.storage = {}


  __shared_states = {}


  def __init__(self, storage_space, wtx):
    path = storage_space.path
    if not path in self.__shared_states:
        self.__shared_states[path]={}
    self.__dict__ = self.__shared_states[path]
    self.path = path
    self.confirmed = ConfirmedTXOStorage(self.path, env=storage_space.env, wtx=wtx)
    self.mempool = self.Interface()
    self.storage_space = storage_space
    self.storage_space.register_txos_storage(self)
      

  def known(self, output_index, rtx):
      return self.confirmed.has(output_index, rtx=rtx) or (output_index in self.mempool)

  def confirm(self, output_index, wtx):
    utxo = self.mempool.storage.pop(output_index)
    self.confirmed.append(utxo, wtx=wtx)

  def apply_tx_get_merkles_and_rollback(self, tx, wtx):
    rollback_inputs = []
    for _i in tx.inputs:
        rollback_inputs.append(self.confirmed.spend(_i, wtx=wtx, return_revert_obj=True))
    for _o in tx.outputs:
        self.confirmed.append(_o, wtx=wtx)
    roots=[self.confirmed.get_commitment_root(rtx=wtx), self.confirmed.get_txo_root(rtx=wtx)]
    for r_i in rollback_inputs:
      self.confirmed.unspend(r_i, wtx=wtx)
    self.confirmed.remove(len(tx.outputs), wtx=wtx)
    return roots

  #TODO bad naming. It should be apply block, or block_transaction
  def apply_tx(self, tx, new_state, wtx):
    rollback_inputs = []
    for _i in tx.inputs:
        if self.storage_space.utxo_index:
          self.storage_space.utxo_index.remove_utxo(_i, wtx=wtx)
        rollback_inputs.append(self.confirmed.spend(_i, return_revert_obj=True, wtx=wtx))
    for _o in tx.outputs:
        if self.storage_space.utxo_index:
          self.storage_space.utxo_index.add_utxo(_o, wtx=wtx)
        self.confirmed.append(_o, wtx=wtx)
        self.mempool.remove(_o)
    self.confirmed.set_state(new_state, wtx=wtx)
    return (rollback_inputs, len(tx.outputs))

  def rollback(self, pruned_inputs, num_of_added_outputs, prev_state, wtx):
    for r_i in pruned_inputs:
      if self.storage_space.utxo_index:
        #r_i[0][2] is serialized txo (0 is txo, 2 is serialized object)
        utxo=IOput()
        utxo.deserialize(r_i[0][2])
        self.storage_space.utxo_index.add_utxo(utxo, wtx=wtx)
      self.confirmed.unspend(r_i, wtx=wtx)
    outputs_for_mempool = self.confirmed.remove(num_of_added_outputs, wtx=wtx)
    for _o in outputs_for_mempool:
      self.mempool[_o.serialized_index]=_o
      if self.storage_space.utxo_index:
        self.storage_space.utxo_index.remove_utxo(_o, wtx=wtx)
    self.confirmed.set_state(prev_state, wtx=wtx)
        

  def find_serialized(self, output_index, rtx):
    if output_index in self.mempool:
      return self.mempool[output_index].serialize()
    else:
      found = self.confirmed.find_wo_deser(output_index, rtx=rtx)
      if found:
        found=found[:-5] #We store ioputs with context, which we cut before relaying
      return found

  def find(self, output_index, rtx):
    if output_index in self.mempool:
      return self.mempool[output_index]
    else:
      return self.confirmed.find(output_index, rtx=rtx)


    





from hashlib import sha256
import progpow
import functools


handler = progpow.ProgPowHandler(max_contexts_num=3)
_progpow_hash = handler.hash

@functools.lru_cache(maxsize=256)
def partial_hash(_bytes):
    seed1 = _bytes
    m1 = sha256()
    m1.update(seed1)
    return m1.digest()

@functools.lru_cache(maxsize=256)
def progpow_hash(header_height, serialized_header_without_nonce, nonce_bytes):
    ph = partial_hash(serialized_header_without_nonce)
    return _progpow_hash(header_height, ph, int.from_bytes(nonce_bytes, "big"))

@functools.lru_cache(maxsize=5)
def seed_hash(header_height):
  return handler.give_seed(header_height)



from leer.core.chains.headers_manager import HeadersManager
from leer.core.chains.blockchain import Blockchain
from leer.core.storage.txos_storage import TXOsStorage
from leer.core.storage.headers_storage import HeadersStorage
from leer.core.storage.blocks_storage import BlocksStorage
from leer.core.storage.excesses_storage import ExcessesStorage
from leer.core.storage.utxo_index_storage import UTXOIndex
from leer.core.storage.mempool_tx import MempoolTx
from leer.core.primitives.block import Block
from leer.core.primitives.header import Header
from leer.core.lubbadubdub.ioput import IOput
from leer.core.lubbadubdub.address import Address
from leer.core.lubbadubdub.transaction import Transaction
from leer.core.hash.progpow import seed_hash as progpow_seed_hash
import base64
from leer.core.utils import DOSException, ObliviousDictionary
from leer.core.primitives.transaction_skeleton import TransactionSkeleton
from time import sleep, time
from uuid import uuid4

from leer.core.storage.storage_space import StorageSpace

from leer.syncer import Syncer
from leer.core.parameters.constants import serialized_genesis_block

from secp256k1_zkp import PrivateKey

import logging
from functools import partial
from ipaddress import ip_address

logger = logging.getLogger("core_loop")

storage_space = None
upload_cache = ObliviousDictionary(sink_delay=600) #tracks what we already sent in recent past to our peers

def arrlen(array):
  return sum([len(el) for el in array])

def init_blockchain(wtx):
  '''
    If blockchain is empty this function will set genesis.
  '''
  genesis = Block(storage_space = storage_space)
  genesis.deserialize(serialized_genesis_block)
  storage_space.headers_manager.set_genesis(genesis.header, wtx=wtx)
  if storage_space.blockchain.current_height(rtx=wtx)<0:
    storage_space.headers_manager.context_validation(genesis.header.hash, rtx=wtx)
    genesis.non_context_verify(rtx=wtx)
    storage_space.blockchain.add_block(genesis, wtx=wtx)
  else:
    storage_space.headers_manager.best_tip = (storage_space.blockchain.current_tip(rtx=wtx), storage_space.blockchain.current_height(rtx=wtx) )
    logger.info("Best header tip from blockchain state %d"%storage_space.headers_manager.best_tip[1])
    #greedy search
    current_tip = storage_space.headers_manager.best_tip[0]
    while True:
      try:
        header = storage_space.headers_storage.get(current_tip, rtx=wtx)
      except KeyError:
        break
      new_current_tip=current_tip
      if len(header.descendants):
        for d in header.descendants:
          dh = storage_space.headers_storage.get(d, rtx=wtx)
          if not dh.invalid:
            new_current_tip = d
            break
      if not new_current_tip == current_tip:
        current_tip=new_current_tip
      else:
        break
    storage_space.headers_manager.best_tip = (current_tip, storage_space.headers_storage.get(current_tip, rtx=wtx).height)
    logger.info("Best header tip after greedy search %d"%storage_space.headers_manager.best_tip[1])


def validate_state(storage_space, rtx):
  '''
    Since writes to different storages (excesses, blocks, txos etc) 
    are not transactional for now, it is possible that previous halt was
    in between of writes. In this case state is (irreversibly) screwed. 
    Cheking it here.
  '''
  if storage_space.blockchain.current_height(rtx=rtx)<1:
    return
  tip = storage_space.blockchain.current_tip(rtx=rtx)
  header = storage_space.headers_storage.get(tip, rtx=rtx)
  last_block_merkles = header.merkles
  state_merkles = [storage_space.txos_storage.confirmed.get_commitment_root(rtx=rtx), \
                   storage_space.txos_storage.confirmed.get_txo_root(rtx=rtx), \
                   storage_space.excesses_storage.get_root(rtx=rtx)]
  try:
    assert last_block_merkles == state_merkles
  except Exception as e:
    logger.error("State is screwed: state merkles are not coinside with last applyed block merkles. Consider full resync.\n %s\n %s\n Block num: %d"%(last_block_merkles, state_merkles, header.height))
    raise e
  

def init_storage_space(config):
  global storage_space
  _path = config["location"]["basedir"]
  storage_space=StorageSpace(_path)
  with storage_space.env.begin(write=True) as wtx:
    hs = HeadersStorage(storage_space, wtx=wtx)
    hm = HeadersManager(storage_space, do_not_check_pow=config.get('do_not_check_pow', False))
    bs = BlocksStorage(storage_space, wtx=wtx)
    es = ExcessesStorage(storage_space, wtx=wtx)
    ts = TXOsStorage(storage_space, wtx=wtx)
    bc = Blockchain(storage_space)
    mptx = MempoolTx(storage_space, config["fee_policy"])
    utxoi = UTXOIndex(storage_space, wtx=wtx)
    init_blockchain(wtx=wtx)
    validate_state(storage_space, rtx=wtx)
  
requests_cache = {"blocks":[], "txouts":[]}
def set_ask_for_blocks_hook(blockchain, message_queue):
  def f(block_hashes):
    if not isinstance(block_hashes, list):
      block_hashes=[block_hashes] #There is only one block
    requests_cache["blocks"]+=block_hashes

  blockchain.ask_for_blocks_hook = f

def set_ask_for_txouts_hook(block_storage, message_queue):
  def f(txouts):
    requests_cache["txouts"]+=txouts

  block_storage.ask_for_txouts_hook = f

def is_ip_port_array(x):
  res = True
  for _ in x:
    try:
      address, port = ip_address(_[0]), int(_[1])
    except:
      res=False
      break
  return res


def set_notify_wallet_hook(blockchain, wallet_message_queue):
    def notify_wallet(reason, *args):
      message={'sender':"Blockchain"}
      #no id: notification
      if reason == "apply":
        message['action'] = "process new block"
        message['tx'] = args[0].serialize()
        message['height'] = args[1]      
      elif reason == "rollback":
        message['action'] = "process rollback"
        message['rollback_object'] = args[0].serialize()
        message['block_height'] = args[1]
      else:
        pass
      wallet_message_queue.put(message)
    storage_space.blockchain.notify_wallet = notify_wallet 

def core_loop(syncer, config):
  message_queue = syncer.queues['Blockchain']
  init_storage_space(config)    

  nodes = {}
  set_ask_for_blocks_hook(storage_space.blockchain, message_queue)
  set_ask_for_txouts_hook(storage_space.blocks_storage, message_queue)
  if config['wallet']:
    set_notify_wallet_hook(storage_space.blockchain, syncer.queues['Wallet'])
  requests = {}
  message_queue.put({"action":"give nodes list reminder"})
  message_queue.put({"action":"check requests cache"})

  #set logging
  default_log_level = logging.INFO;
  if "logging" in config:#debug, info, warning, error, critical
    loglevels = { "debug":logging.DEBUG, "info":logging.INFO, "warning":logging.WARNING, "error":logging.ERROR, "critical":logging.CRITICAL}
    if "base" in config["logging"] and config["logging"]["base"] in loglevels:
      logger.setLevel(loglevels[config["logging"]["base"]])
    if "core" in config["logging"] and config["logging"]["core"] in loglevels:
      #its ok to rewrite
      logger.setLevel(loglevels[config["logging"]["core"]])

  def get_new_address(timeout=2.5): #blocking
    _id = str(uuid4())
    syncer.queues['Wallet'].put({'action':'give new address', 'id':_id, 'sender': "Blockchain"})
    result = None
    start_time=time()
    while True:
      put_back = [] #We wait for specific message, all others will wait for being processed
      while not message_queue.empty():
        message = message_queue.get()
        if (not 'id' in message)  or (not message['id']==_id):
          put_back.append(message)
          continue
        result = message['result']
        break
      for message in put_back:
        message_queue.put(message)
      if result:
        break
      sleep(0.01)
      if time()-start_time>timeout:
        raise Exception("get_new_address timeout: probably wallet has collapsed or not running")      
    if result=='error':
      raise Exception("Can not get_new_address: error on wallet side")      
    address = Address()
    logger.info("Receiving address %s (len %d)"%( result, len(result)))
    address.deserialize_raw(result)
    return address

  mining_address = None #Will be initialised at first ask

  def send_message(destination, message):
    logger.debug("Sending message to %s:\t\t %s"%(str(destination), str(message)))
    if not 'id' in message:
      message['id'] = uuid4()
    if not 'sender' in message:
      message['sender'] = "Blockchain"
    syncer.queues[destination].put(message)

  def send_to_nm(message):
    send_message("NetworkManager", message)

  notification_cache = {}
  def notify(key, value, timestamp=None):
    if (key in notification_cache) and (notification_cache[key]['value'] == value) and (time()-notification_cache[key]['timestamp'])<5:
      return #Do not spam notifications with the same values
    message = {}
    message['id'] = uuid4()
    message['sender'] = "Blockchain"
    if not timestamp:
      timestamp = time()
    message['time'] = timestamp
    message['action']="set"
    message['key']=key
    message['value']=value
    syncer.queues["Notifications"].put(message)
    notification_cache[key] = {'value':value, 'timestamp':timestamp}

  logger.debug("Start of core loop")
  with storage_space.env.begin(write=True) as rtx: #Set basic chain info, so wallet and other services can start work
    notify("blockchain height", storage_space.blockchain.current_height(rtx=rtx))
    notify("best header", storage_space.headers_manager.best_header_height)         
  while True:
    sleep(0.05)
    put_back_messages = []
    notify("core workload", "idle")
    while not message_queue.empty():
      message = message_queue.get()
      if 'time' in message and message['time']>time(): # delay this message
        put_back_messages.append(message)
        continue
      if (('result' in message) and message['result']=="processed") or \
         (('result' in message) and message['result']=="set") or \
         (('action' in message) and message['action']=="give nodes list reminder") or \
         (('action' in message) and message['action']=="check requests cache") or \
         (('action' in message) and message['action']=="take nodes list") or \
         (('result' in message) and is_ip_port_array(message['result'])):
        logger.debug("Processing message %s"%message)
      else:
        if 'action' in message:
          logger.info("Processing message `%s`"%message['action'])
        else:
          logger.info("Processing message %s"%message)  
      if not 'action' in message: #it is response
        if message['id'] in requests: # response is awaited
          if requests[message['id']]=="give nodes list":
            requests.pop(message['id'])
            message_queue.put({"action":"take nodes list", "nodes":message["result"]})
        else:
          pass #Drop
        continue
      try:
        if message["action"] == "take the headers":
          notify("core workload", "processing new headers")
          with storage_space.env.begin(write=True) as wtx:
            process_new_headers(message, nodes[message["node"]], send_message, wtx, notify=partial(notify, "best header"))
          notify("best header", storage_space.headers_manager.best_header_height)         
        if message["action"] == "take the blocks":
          notify("core workload", "processing new blocks")
          with storage_space.env.begin(write=True) as wtx:
            initial_tip = storage_space.blockchain.current_tip(rtx=wtx)
            process_new_blocks(message, notify=partial(notify, "blockchain height"), wtx=wtx)
            after_tip = storage_space.blockchain.current_tip(rtx=wtx)
            notify("blockchain height", storage_space.blockchain.current_height(rtx=wtx))         
            if not after_tip==initial_tip:
              notify_all_nodes_about_new_tip(nodes, send_to_nm, rtx=wtx, _except=[message["node"]]) 
            look_forward(nodes, send_to_nm, rtx=wtx)       
        if message["action"] == "take the txos":
          notify("core workload", "processing new txos")
          with storage_space.env.begin(write=True) as wtx:
            process_new_txos(message, wtx=wtx)
            #After downloading new txos some blocs may become downloaded
            notify("blockchain height", storage_space.blockchain.current_height(rtx=wtx)) 
            look_forward(nodes, send_to_nm, rtx=wtx)          
        if message["action"] == "give blocks":
          notify("core workload", "giving blocks")
          with storage_space.env.begin(write=False) as rtx:
            process_blocks_request(message, send_message, rtx=rtx)
        if message["action"] == "give next headers":
          notify("core workload", "giving headers")
          with storage_space.env.begin(write=False) as rtx:
            process_next_headers_request(message, send_message, rtx=rtx)
        if message["action"] == "give txos":
          notify("core workload", "giving txos")
          with storage_space.env.begin(write=False) as rtx:
            process_txos_request(message, send_message, rtx=rtx)
        if message["action"] == "find common root":
          with storage_space.env.begin(write=False) as rtx:
            process_find_common_root(message, send_message, rtx)
        if message["action"] == "find common root response":
          with storage_space.env.begin(write=False) as rtx:
            process_find_common_root_response(message, nodes[message["node"]], send_message, rtx=rtx)
        if message["action"] == "give TBM transaction":
          notify("core workload", "giving mempool tx")
          with storage_space.env.begin(write=False) as rtx:
            process_tbm_tx_request(message, send_message, rtx)
        if message["action"] == "take TBM transaction":
          notify("core workload", "processing mempool tx")
          with storage_space.env.begin(write=False) as rtx:
            process_tbm_tx(message, send_to_nm, nodes, rtx=rtx)
        if message["action"] == "give tip height":
          with storage_space.env.begin(write=False) as rtx:
            _ch=storage_space.blockchain.current_height(rtx=rtx)
            send_message(message["sender"], {"id": message["id"], "result": _ch})
          notify("blockchain height", _ch)      
        if message["action"] == "take tip info":
          if not message["node"] in nodes:
            nodes[message["node"]]={'node':message["node"]}
          with storage_space.env.begin(write=False) as rtx:
            process_tip_info(message, nodes[message["node"]], rtx=rtx, send=send_to_nm)
      except DOSException as e:
        logger.info("DOS Exception %s"%str(e))
        #raise e #TODO send to NM
      except Exception as e:
        raise e

      if message["action"] == "give block info":
        notify("core workload", "reading block info")
        try:
          with storage_space.env.begin(write=False) as rtx:
            block_info = compose_block_info(message["block_num"], rtx=rtx)
          send_message(message["sender"], {"id": message["id"], "result":block_info})
        except Exception as e:
          send_message(message["sender"], {"id": message["id"], "result":"error", "error":str(e)})

      if message["action"] == "give block template":
        notify("core workload", "generating block template")
        try:
          if not mining_address:
            mining_address = get_new_address()
          with storage_space.env.begin(write=True) as wtx:
            if "mining" in config and "conditions" in config["mining"]:
              if "connected" in config["mining"]["conditions"]:
                if not len(nodes):
                  raise Exception("Cant start mining with zero connections")
              if "synced_headers" in config["mining"]["conditions"]:
                best_block, best_header = storage_space.blockchain.current_height(rtx=wtx),\
                                          storage_space.headers_manager.best_header_height
                if best_block<best_header:
                  raise Exception("Cant start mining while best block %d worse than best known header %d"%(best_block, best_header))
              if "synced_advertised" in config["mining"]["conditions"]:
                best_block, best_advertised_height = storage_space.blockchain.current_height(rtx=wtx),\
                                                     max([nodes[node]["height"] for node in nodes if "height" in nodes[node]])
                if best_block<best_advertised_height:
                  raise Exception("Cant start mining while best block %d worse than best advertised block %d"%(best_block, best_advertised_height))
            block = storage_space.mempool_tx.give_block_template(mining_address, wtx=wtx)
          ser_head = block.header.serialize()
          send_message(message["sender"], {"id": message["id"], "result":ser_head})
        except Exception as e:
          send_message(message["sender"], {"id": message["id"], "result":"error", "error":str(e)})
          logger.error("Can not generate block `%s`"%(str(e)), exc_info=True)
      if message["action"] == "give mining work":
        notify("core workload", "generating block template")
        try:
          if not mining_address:
            mining_address = get_new_address()
          with storage_space.env.begin(write=True) as wtx:
            if "mining" in config and "conditions" in config["mining"]:
              if "connected" in config["mining"]["conditions"]:
                if not len(nodes):
                  raise Exception("Cant start mining with zero connections")
              if "synced_headers" in config["mining"]["conditions"]:
                best_block, best_header = storage_space.blockchain.current_height(rtx=wtx),\
                                          storage_space.headers_manager.best_header_height
                if best_block<best_header:
                  raise Exception("Cant start mining while best block %d worse than best known header %d"%(best_block, best_header))
              if "synced_advertised" in config["mining"]["conditions"]:
                best_block, best_advertised_height = storage_space.blockchain.current_height(rtx=wtx),\
                                                     max([nodes[node]["height"] for node in nodes if "height" in nodes[node]])
                if best_block<best_advertised_height:
                  raise Exception("Cant start mining while best block %d worse than best advertised block %d"%(best_block, best_advertised_height))
            partial_header_hash, target, height = storage_space.mempool_tx.give_mining_work(mining_address, wtx=wtx)
          seed_hash = progpow_seed_hash(height)
          send_message(message["sender"], {"id": message["id"], 
              "result":{'partial_hash':partial_header_hash.hex(), 
                        'seed_hash':seed_hash.hex(),
                        'target':target.hex(),
                        'height':height
                       }})
        except Exception as e:
          send_message(message["sender"], {"id": message["id"], "result":"error", "error":str(e)})
          logger.error("Can not generate work `%s`"%(str(e)), exc_info=True)
      if message["action"] == "take solved block template":
        notify("core workload", "processing solved block")
        try:
          with storage_space.env.begin(write=True) as wtx:
            initial_tip = storage_space.blockchain.current_tip(rtx=wtx)
            header = Header()
            header.deserialize(message["solved template"])
            solved_block = storage_space.mempool_tx.get_block_by_header_solution(header)
            storage_space.headers_manager.add_header(solved_block.header, wtx=wtx)
            storage_space.headers_manager.context_validation(solved_block.header.hash, rtx=wtx)
            solved_block.non_context_verify(rtx=wtx)
            storage_space.blockchain.add_block(solved_block, wtx=wtx)
            after_tip = storage_space.blockchain.current_tip(rtx=wtx)
            our_height = storage_space.blockchain.current_height(rtx=wtx)
            best_known_header = storage_space.headers_manager.best_header_height
            if not after_tip==initial_tip:
              notify_all_nodes_about_new_tip(nodes, send_to_nm, rtx=wtx)
          send_message(message["sender"], {"id": message["id"], "result": "Accepted"})
          notify("best header", best_known_header)
          notify("blockchain height", our_height)
        except Exception as e:
          logger.error("Wrong block solution %s"%str(e))
          send_message(message["sender"], {"id": message["id"], "error": str(e), 'result':'error'})
      if message["action"] == "take mining work":
        notify("core workload", "processing mining work")
        try:
          nonce, partial_work = message['nonce'], message['partial_hash']
          mp = storage_space.mempool_tx
          block_template =  mp.work_block_assoc[partial_work]
          block_template.header.nonce = nonce
          solved_block = block_template 
          header = solved_block.header
          with storage_space.env.begin(write=True) as wtx:
            if header.height <= storage_space.blockchain.current_height(rtx=wtx):
              send_message(message["sender"], {"id": message["id"], "result": "Stale"})
              logger.error("Stale work submitted: height %d"%(header.height))
              continue
            initial_tip = storage_space.blockchain.current_tip(rtx=wtx)
            storage_space.headers_manager.add_header(solved_block.header, wtx=wtx)
            storage_space.headers_manager.context_validation(solved_block.header.hash, rtx=wtx)
            solved_block.non_context_verify(rtx=wtx)
            storage_space.blockchain.add_block(solved_block, wtx=wtx)
            after_tip = storage_space.blockchain.current_tip(rtx=wtx)
            our_height = storage_space.blockchain.current_height(rtx=wtx)
            best_known_header = storage_space.headers_manager.best_header_height
            if not after_tip==initial_tip:
              notify_all_nodes_about_new_tip(nodes, send_to_nm, rtx=wtx)
          send_message(message["sender"], {"id": message["id"], "result": "Accepted"})
          notify("best header", best_known_header)
          notify("blockchain height", our_height)
        except Exception as e:
          logger.error("Wrong submitted work %s"%str(e))
          send_message(message["sender"], {"id": message["id"], "error": str(e), 'result':'error'})
      if message["action"] == "give synchronization status":
        with storage_space.env.begin(write=False) as rtx:
          our_height = storage_space.blockchain.current_height(rtx=rtx)
        best_known_header = storage_space.headers_manager.best_header_height
        try:
          best_advertised_height = max([nodes[node]["height"] for node in nodes if "height" in nodes[node]])
        except:
          best_advertised_height = None
        send_message(message["sender"], {"id": message["id"], 
                                         "result": {'height': our_height, 
                                                    'best_known_header': best_known_header,
                                                    'best_advertised_height': best_advertised_height}})
        notify("best header", best_known_header)
        notify("blockchain height", our_height)
        notify("best advertised height", best_advertised_height)

      if message["action"] == "add tx to mempool":
        notify("core workload", "processing local transaction")
        response = {"id": message["id"]}
        #deserialization
        try:
          ser_tx = message["tx"]
          tx = Transaction(txos_storage = storage_space.txos_storage, excesses_storage = storage_space.excesses_storage)
          with storage_space.env.begin(write=False) as rtx:            
            tx.deserialize(ser_tx, rtx)
            storage_space.mempool_tx.add_tx(tx, rtx=rtx)
            tx_skel = TransactionSkeleton(tx=tx)
            notify_all_nodes_about_tx(tx_skel.serialize(rich_format=True, max_size=40000), nodes, send_to_nm, _except=[], mode=1)
          response['result']="generated"
        except Exception as e:
          response['result'] = 'error'
          response['error'] = str(e)
          logger.error("Problem in tx: %s"%str(e))
        send_message(message["sender"], response)


      #message from core_loop
      if message["action"] == "check txouts download status":
        txos = message["txos_hashes"]
        to_be_downloaded = []
        with storage_space.env.begin(write=True) as rtx:
          for txo in txos:
            if not storage_space.txos_storage.known(txo, rtx=rtx):
              to_be_downloaded.append(txo)
        if not to_be_downloaded:
          continue #We are good, txouts are already downloaded
        already_asked_nodes = message["already_asked_nodes"]
        asked = False
        for node_params in nodes:
          node = nodes[node_params]
          if node in already_asked_nodes:
            continue
          already_asked_nodes += [node]
          send_to_nm({"action":"give txos",
                                               "txos_hashes": b"".join(to_be_downloaded), 
                                               "num": len(to_be_downloaded), 
                                               "id":str(uuid4()), "node":node_params })
          new_message = {"action": "check txouts download status", "txos_hashes":to_be_downloaded,
                         "already_asked_nodes": already_asked_nodes, "id": str(uuid4()),
                         "time": int(time()+30) }
          asked = True
          put_back_messages.append(new_message)
          break
        if not asked: #We already asked all applicable nodes
          message["time"]=int(time())+600
          message["already_asked_nodes"] = []
          put_back_messages.append(message) # we will try to ask again in an hour

      #message from core_loop
      if message["action"] == "check blocks download status":
        block_hashes = message["block_hashes"]
        to_be_downloaded = []
        lowest_height=1e10
        with storage_space.env.begin(write=True) as rtx:
          for block_hash in block_hashes:
            if storage_space.blocks_storage.has(block_hash, rtx=rtx):
              continue #We are good, block already downloaded          
            if not block_hash in storage_space.blockchain.awaited_blocks:
              continue #For some reason we don't need this block anymore
            to_be_downloaded.append(block_hash)
            block_height = storage_space.headers_storage.get(block_hash, rtx=rtx).height
            if block_height<lowest_height:
              lowest_height = block_height
        if not len(to_be_downloaded):
          continue #We are good, blocks are already downloaded
        already_asked_nodes = message["already_asked_nodes"]
        asked = False
        for node_params in nodes:
          node = nodes[node_params]
          if node in already_asked_nodes:
            continue
          if (not "height" in node) or node["height"] < lowest_height:
            continue
          already_asked_nodes += [node]
          send_to_nm({"action":"give blocks",  "block_hashes": bytes(b"".join(block_hashes)), 'num': len(block_hashes), "id":str(uuid4()), "node":node_params })
          new_message = {"action": "check blocks download status", "block_hashes":to_be_downloaded,
                         "already_asked_nodes": already_asked_nodes, "id": str(uuid4()),
                         "time": int(time()+30) }
          asked = True
          put_back_messages.append(new_message)
          break
        if not asked: #We already asked all applicable nodes
          message["time"]=int(time())+600
          message["already_asked_nodes"] = []
          put_back_messages.append(message) # we will try to ask again in an hour


      if message["action"] == "take nodes list":
        for node in message["nodes"]:
          if not node in nodes: #Do not overwrite
            nodes[node]={"node":node}
        disconnected_nodes = []
        for existing_node in nodes:
          if not existing_node in message["nodes"]:
            disconnected_nodes.append(existing_node)
        for dn in disconnected_nodes:
          nodes.pop(dn)
        

      if message["action"] == "give nodes list reminder":
        _id = str(uuid4())
        send_to_nm({"action":"give intrinsic nodes list", "sender":"Blockchain", "id":_id})
        requests[_id] = "give nodes list"
        put_back_messages.append({"action": "give nodes list reminder", "time":int(time())+3} )

      if message["action"] == "stop":
        logger.info("Core loop stops")
        return

      if message["action"] == "check requests cache":
        put_back_messages.append({"action": "check requests cache", "time":int(time())+5} )
        for k in requests_cache:
          if not len(requests_cache[k]):
            continue
          copy = list(set(requests_cache[k]))
          if k=="blocks":
            chunk_size=20
            while len(copy):
              request, copy = copy[:chunk_size], copy[chunk_size:]
              new_message = {"action": "check blocks download status", "block_hashes":request,
                            "already_asked_nodes": [], "id": str(uuid4()),
                            "time": -1 }
              message_queue.put(new_message)
            requests_cache[k] = []
          if k=="txouts":
            chunk_size=30
            while len(copy):
              request, copy = copy[:chunk_size], copy[chunk_size:]
              new_message = {"action": "check txouts download status", "txos_hashes": request,
                           "already_asked_nodes": [], "id": str(uuid4()),
                           "time": -1 }
              message_queue.put(new_message)
            requests_cache[k] = []

    for _message in put_back_messages:
      message_queue.put(_message)

    try:
      with storage_space.env.begin(write=True) as rtx:
        check_sync_status(nodes, send_to_nm, rtx=rtx)
      try:
        best_advertised_height = max([nodes[node]["height"] for node in nodes if "height" in nodes[node]])
      except:
          best_advertised_height = None
      notify("best advertised height", best_advertised_height)
    except Exception as e:
      logger.error(e)

def look_forward(nodes, send_to_nm, rtx):
  if storage_space.headers_manager.best_header_height < storage_space.blockchain.current_height(rtx=rtx)+100:
    for node_index in nodes:
      node = nodes[node_index]
      if ('height' in node) and (node['height']>storage_space.headers_manager.best_header_height):
        our_tip_hash = storage_space.blockchain.current_tip(rtx=rtx)
        send_find_common_root(storage_space.headers_storage.get(our_tip_hash,rtx=rtx), node['node'], send = send_to_nm)
        break




def process_new_headers(message, node_info, send_message, wtx, notify=None):
  dupplication_header_dos = False #TODO grammatical typo?
  try:
    serialized_headers = message["headers"]
    num = message["num"]
    header = None
    for i in range(num):
      header = Header()
      serialized_headers = header.deserialize_raw(serialized_headers)
      if not storage_space.headers_storage.has(header.hash, rtx=wtx):
        storage_space.headers_manager.add_header(header, wtx=wtx)
        if notify and not i%20:
          notify(storage_space.headers_manager.best_header_height)
      else:
        dupplication_header_dos = True
    if ("common_root" in node_info) and ("long_reorganization" in node_info["common_root"]):
        if storage_space.headers_manager.get_best_tip()[0] == header.hash:
          #not reorg anymore
          node_info["common_root"].pop("long_reorganization", None)          
          our_tip_hash = storage_space.blockchain.current_tip(rtx=wtx)
          storage_space.blockchain.update(wtx=wtx, reason="downloaded new headers")
          send_tip_info(node_info=node_info, send = partial(send_message, "NetworkManager"), our_tip_hash=our_tip_hash, rtx=wtx)
        elif node_info["common_root"]["long_reorganization"]==header.height:
           request_num = min(256, node_info["height"]-storage_space.headers_storage.get(node_info["common_root"]["root"], rtx=wtx).height) 
           send_next_headers_request(header.hash, 
                                request_num,
                                message["node"], send = partial(send_message, "NetworkManager") )
           if node_info["height"]-header.height>request_num:
             node_info["common_root"]["long_reorganization"] = header.height+request_num
           else:
             node_info["common_root"].pop("long_reorganization", None) 
    storage_space.blockchain.update(wtx=wtx, reason="downloaded new headers")
  except Exception as e:
    raise DOSException() #TODO add info

def process_new_blocks(message, wtx, notify=None):
  try:
    serialized_blocks = message["blocks"]
    num = message["num"]
    prev_not = time()
    for i in range(num):
      block = Block(storage_space=storage_space)
      serialized_blocks = block.deserialize_raw(serialized_blocks)
      storage_space.blockchain.add_block(block, wtx=wtx, no_update=True)
      if notify:
        if time()-prev_not>15:
          storage_space.blockchain.update(wtx=wtx, reason="downloaded new blocks")
          notify(storage_space.blockchain.current_height(rtx=wtx))
          prev_not = time()
    storage_space.blockchain.update(wtx=wtx, reason="downloaded new blocks")
  except Exception as e:
    raise e
    raise DOSException() #TODO add info

def process_new_txos(message, wtx):
  try:
    serialized_utxos = bytes(message["txos"])
    txos_hashes = bytes(message["txos_hashes"])
    num = message["num"]
    #txos_hashes = [txos_hashes[i*65:(i+1)*65)] for i in range(0,num)]
    txos_lengths = message["txos_lengths"]
    #TODO we should use txos_hashes and txos_lengths to proccess only unknown txos
    for i in range(num):
      utxo = IOput()
      serialized_utxos = utxo.deserialize_raw(serialized_utxos)
      storage_space.txos_storage.mempool[utxo.serialized_index]=utxo
    storage_space.blockchain.update(wtx=wtx, reason="downloaded new txos")
  except Exception as e:
    raise DOSException() #TODO add info

def process_blocks_request(message, send_message, rtx):
  num = message["num"]
  _hashes = message["block_hashes"]
  _hashes = [_hashes[i*32:(i+1)*32] for i in range(num)]
  serialized_blocks = []
  blocks_hashes = [] 
  for _hash in _hashes:
    if not storage_space.blocks_storage.has(_hash, rtx=rtx):
      continue
    try:
      serialized_block = storage_space.blocks_storage.get(_hash, rtx=rtx).serialize(rtx=rtx, rich_block_format=True)
    except KeyError:
      #Some outputs were pruned
      serialized_block = storage_space.blocks_storage.get(_hash, rtx=rtx).serialize(rtx=rtx, rich_block_format=False)
    if arrlen(serialized_blocks)+len(serialized_block)<60000:
      serialized_blocks.append(serialized_block)
      blocks_hashes.append(_hash)
    else:
        send_blocks(partial(send_message, message['sender']), \
                     blocks = serialized_blocks, \
                     hashes = blocks_hashes, \
                     node = message["node"], \
                     _id=message['id'])
        serialized_blocks=[serialized_block]
        blocks_hashes = [_hash]
  send_blocks(partial(send_message, message['sender']), \
              blocks = serialized_blocks, \
              hashes = blocks_hashes, \
              node = message["node"], \
              _id=message['id'])

def send_next_headers_request(from_hash, num, node, send):
  send({"action":"give next headers", "num":num, "from":from_hash, 
                                       "id" : str(uuid4()), "node": node  })

def process_next_headers_request(message, send_message, rtx):
  from_hash = message["from"]
  num = message["num"]
  num = 1024 if num>1024 else num
  try:
    header = storage_space.headers_storage.get(from_hash, rtx=rtx)
  except KeyError:
    return #unknown hash
  current_tip  = storage_space.blockchain.current_tip(rtx=rtx)
  current_height  = storage_space.blockchain.current_height(rtx=rtx)
  if not storage_space.headers_manager.find_ancestor_with_height(current_tip, header.height, rtx=rtx) == from_hash:
    return
    ''' Counter-node is not in our main chain. 
        We will not feed it (actually we just are not sure what we should send here)
    '''
  last_to_send_height = header.height+num
  last_to_send_height = current_height if last_to_send_height>current_height else last_to_send_height
  last_to_send = storage_space.headers_manager.find_ancestor_with_height(current_tip, last_to_send_height, rtx=rtx)
  headers_hashes = storage_space.headers_manager.get_subchain(from_hash, last_to_send, rtx=rtx)

  serialized_headers = []
  out_headers_hashes = []
  for _hash in headers_hashes:
    if not storage_space.headers_storage.has(_hash, rtx=rtx):
      continue
    serialized_header = storage_space.headers_storage.get(_hash, rtx=rtx).serialize()
    if arrlen(serialized_headers)+len(serialized_header)<60000:
      serialized_headers.append(serialized_header)
      out_headers_hashes.append(_hash)
    else:
      send_headers(partial(send_message, message['sender']), \
                   headers = serialized_headers, \
                   hashes = out_headers_hashes,\
                   node = message["node"], \
                   _id=message['id'])
      serialized_headers=[serialized_header]
      out_headers_hashes = [_hash]
  send_headers(partial(send_message, message['sender']), \
                   headers = serialized_headers,\
                   hashes = out_headers_hashes,\
                   node = message["node"], \
                   _id=message['id'])

def process_txos_request(message, send_message, rtx):
  num = message["num"]
  _hashes = message["txos_hashes"]
  _hashes = [bytes(_hashes[i*65:(i+1)*65]) for i in range(num)]
  serialized_txos = []
  txos_hashes = []
  for _hash in _hashes:
    try:
      serialized_txo = storage_space.txos_storage.find_serialized(_hash, rtx=rtx)
    except KeyError:
      continue
    len_txos_hashes = 65*len(serialized_txos)
    len_txos_lens = 2*len(serialized_txos)
    if arrlen(serialized_txos)+len(serialized_txo)+len_txos_hashes+len_txos_lens<64000:
      serialized_txos.append(serialized_txo)
      txos_hashes.append(_hash)
    else:
      send_txos(partial(send_message, message['sender']), \
                   txos = serialized_txos,\
                   hashes = txos_hashes,\
                   node = message["node"], \
                   _id=message['id'])
      serialized_txos=[serialized_txo]
      txos_hashes = [_hash]
  send_txos(partial(send_message, message['sender']), \
                   txos = serialized_txos,\
                   hashes = txos_hashes,\
                   node = message["node"], \
                   _id=message['id'])

def send_tip_info(node_info, send, rtx, our_tip_hash=None ):
  our_height = storage_space.blockchain.current_height(rtx=rtx)
  our_tip_hash = our_tip_hash if our_tip_hash else storage_space.blockchain.current_tip(rtx=rtx)
  our_prev_hash = storage_space.headers_storage.get(our_tip_hash, rtx=rtx).prev
  our_td = storage_space.headers_storage.get(our_tip_hash, rtx=rtx).total_difficulty

  send({"action":"take tip info", "height":our_height, "tip":our_tip_hash, "prev_hash":our_prev_hash, "total_difficulty":our_td, "id":uuid4(), "node": node_info["node"] })
  node_info["sent_tip"]=our_tip_hash
  node_info["last_send"] = time()

def process_tip_info(message, node_info, send, rtx):
  # another node (referenced as counter-Node below) asks us for our best tip and also provides us information about his
  if ('common_root' in node_info) and ('worst_nonmutual' in node_info['common_root']):
    return #Ignore new tips while detect common_root
  node = message["node"]
  height = message["height"]
  tip_hash = message["tip"]
  prev_hash = message["prev_hash"]
  total_difficulty = message["total_difficulty"]
  our_tip_hash = storage_space.blockchain.current_tip(rtx=rtx)

  if (not "sent_tip" in node_info) or \
     (not node_info["sent_tip"]==our_tip_hash) or \
     (not "last_send" in node_info) or \
     (time() - node_info["last_send"]>300):
    send_tip_info(node_info=node_info, send = send, our_tip_hash=our_tip_hash, rtx=rtx)
  node_info.update({"node":node, "height":height, "tip_hash":tip_hash, 
                    "prev_hash":prev_hash, "total_difficulty":total_difficulty, 
                    "last_update":time()})
  if (height > storage_space.blockchain.current_height(rtx=rtx)) and (total_difficulty > storage_space.headers_storage.get(our_tip_hash, rtx=rtx).total_difficulty):
    #Now there are two options: better headers are unknown or headers are known, but blocks are unknown or bad
    if not storage_space.headers_storage.has(tip_hash, rtx=rtx):
      send_find_common_root(storage_space.headers_storage.get(our_tip_hash, rtx=rtx), node, send = send)
      #TODO check prev hash first
    else: #header is known
      header = storage_space.headers_storage.get(tip_hash, rtx=rtx)
      #TODO do we need to check for connection to genesis?
      common_root =  storage_space.headers_manager.find_bifurcation_point(tip_hash, our_tip_hash, rtx=rtx)
      if header.invalid:
        return #Nothing interesting, counter-node is on wrong chain
      if storage_space.blocks_storage.has(tip_hash, rtx=rtx):
        if storage_space.blocks_storage.get(tip_hash, rtx=rtx).invalid:
          return #Nothing interesting, counter-node is on wrong chain
      #download blocks
      blocks_to_download = []
      for _block_hash in storage_space.headers_manager.get_subchain(common_root, tip_hash, rtx=rtx):
        if not storage_space.blocks_storage.has(_block_hash, rtx=rtx):
          blocks_to_download.append(_block_hash)
        else:
          storage_space.blocks_storage.is_block_downloaded(_block_hash, rtx=rtx)
        if len(blocks_to_download)*32>40000: #Too big for one message
          break
      if len(blocks_to_download):
        send({"action":"give blocks",  
            "block_hashes": b"".join(blocks_to_download),
            'num': len(blocks_to_download), "id":str(uuid4()), "node":message["node"] })
      

def send_find_common_root(from_header, node, send):
  send(
    {"action":"find common root", "serialized_header": from_header.serialize(), 
     "id":str(uuid4()), 
     "node": node })


UNKNOWN, INFORK, MAINCHAIN, ISOLATED = 0, 1, 2, 3

def process_find_common_root(message, send_message, rtx):
  try:
    serialized_header = message["serialized_header"]
    header = Header()
    header.deserialize_raw(serialized_header)
  except:
    raise DOSException()
  result = []
  for pointer in [header.hash]+header.popow.pointers:
    if not storage_space.headers_storage.has(pointer, rtx=rtx):
      result.append(UNKNOWN)
      continue
    ph = storage_space.headers_storage.get(pointer, rtx=rtx)
    if not ph.connected_to_genesis:
      result.append(ISOLATED)
      continue
    if storage_space.headers_manager.find_ancestor_with_height(storage_space.blockchain.current_tip(rtx=rtx), ph.height, rtx=rtx) == pointer:
      result.append(MAINCHAIN)
      continue
    result.append(INFORK)

  send_message(message['sender'], \
     {"action":"find common root response", "header_hash":header.hash,
      "flags_num": len(result), 
      "known_headers": b"".join([i.to_bytes(1,"big") for i in result]), 
      "id":message['id'], "node": message["node"] })

def process_find_common_root_response(message, node_info, send_message, rtx):
  logger.info("Processing of fcrr")
  header_hash = message["header_hash"]
  result = [int(i) for i in message["known_headers"]]
  try:
    header = storage_space.headers_storage.get(header_hash, rtx=rtx)
  except KeyError:
    raise DOSException()
  root_found = False
  if not "common_root" in node_info:
    node_info["common_root"]={}

  for index, pointer in enumerate([header.hash]+header.popow.pointers):
    if result[index] in [MAINCHAIN]:
        node_info["common_root"]["best_mutual"]=pointer
        best_mutual_height = storage_space.headers_storage.get(node_info["common_root"]["best_mutual"], rtx=rtx).height
        break
    else:
      node_info["common_root"]["worst_nonmutual"]=pointer
  if (not "worst_nonmutual" in node_info["common_root"]):
    #we are behind
    node_info["common_root"]["root"] = header_hash
    root_found = True
    node_info["common_root"].pop("try_num", None)
  if (not "best_mutual" in node_info["common_root"]):
    # genesis should always be mutual
    return
  logger.info(node_info)
  if not root_found:
    h1,h2 = storage_space.headers_storage.get(node_info["common_root"]["worst_nonmutual"], rtx=rtx).height,\
            storage_space.headers_storage.get(node_info["common_root"]["best_mutual"], rtx=rtx).height
    if h1==h2+1:
      root_found = True
      node_info["common_root"].pop("try_num", None)
      node_info["common_root"]["root"] = node_info["common_root"]["best_mutual"]
    else:
      if not "try_num" in node_info["common_root"]:
        node_info["common_root"]["try_num"]=0
      node_info["common_root"]["try_num"]+=1
      send_find_common_root(storage_space.headers_storage.get(node_info["common_root"]["worst_nonmutual"], rtx=rtx), message['node'],\
                          send = partial(send_message, "NetworkManager") )
      if node_info["common_root"]["try_num"]>5:
        pass #TODO we shoould try common root not from worst_nonmutual but at the middle between worst_nonmutual and best_mutual (binary search)
  try:
    height, total_difficulty = node_info['height'],node_info['total_difficulty']
  except KeyError:
    return
  logger.info((height, storage_space.headers_manager.best_header_height, total_difficulty , storage_space.headers_manager.best_header_total_difficulty(rtx=rtx)))
  if root_found:
    node_info["common_root"].pop("worst_nonmutual", None)
    node_info["common_root"].pop("best_mutual", None)
    common_root_height = storage_space.headers_storage.get(node_info["common_root"]["root"], rtx=rtx).height
    if (height > storage_space.headers_manager.best_header_height) and (total_difficulty > storage_space.headers_manager.best_header_total_difficulty(rtx=rtx)):
      request_num = min(256, height-common_root_height)
      send_next_headers_request(node_info["common_root"]["root"], 
                                request_num,
                                message["node"], send = partial(send_message, "NetworkManager") )
      if height-common_root_height>request_num:
        our_tip = storage_space.headers_storage.get(storage_space.blockchain.current_tip(rtx=rtx), rtx=rtx)
        if our_tip.hash != node_info["common_root"]["root"]: #It's indeed reorg
            node_info["common_root"]["long_reorganization"]= storage_space.headers_storage.get(node_info["common_root"]["root"], rtx=rtx).height+request_num



def  process_tbm_tx_request(message, send_message, rtx):
  tx_skel = storage_space.mempool_tx.give_tx_skeleton()
  tx = storage_space.mempool_tx.give_tx()
  serialized_tx_skel = tx_skel.serialize(rich_format=True, max_size=60000, rtx=rtx, full_tx=tx)
  send_message(message['sender'], \
     {"action":"take TBM transaction", "tx_skel": serialized_tx_skel, "mode": 0,
      "id":message['id'], 'node': message["node"] })

def  process_tbm_tx(message, send, nodes, rtx):
  try:
    initial_tbm = storage_space.mempool_tx.give_tx()
    tx_skel = TransactionSkeleton()
    tx_skel.deserialize_raw(message['tx_skel'], storage_space = storage_space)
    storage_space.mempool_tx.add_tx(tx_skel, rtx=rtx)
    final_tbm = storage_space.mempool_tx.give_tx()
    if not message["mode"]==0: #If 0 it is response to our request
      if (not initial_tbm) or (not str(initial_tbm.serialize())==str(final_tbm.serialize())):
        notify_all_nodes_about_tx(message['tx_skel'], nodes, send, _except=[message["node"]])
  except Exception as e:
    print(e)
    pass


def compose_block_info(block_num, rtx):
  ct = storage_space.blockchain.current_tip(rtx=rtx)
  ch = storage_space.blockchain.current_height(rtx=rtx)
  if block_num>ch:
    raise Exception("Unknown block")
  target_hash = ct
  if block_num<ch:
    target_hash = storage_space.headers_manager.find_ancestor_with_height(ct, block_num, rtx=rtx)
  block = storage_space.blocks_storage.get(target_hash, rtx=rtx)
  result = {'hash':target_hash.hex()}
  result['target']=float(block.header.target)
  result['supply']=block.header.supply
  result['timestamp']=block.header.timestamp
  result['height'] = block.header.height
  result['inputs']=[]
  result['outputs']=[]
  for i in block.transaction_skeleton.input_indexes:
    index=i.hex()
    address=storage_space.txos_storage.find(i, rtx=rtx).address.to_text()
    result['inputs'].append((index, address))
  for o in block.transaction_skeleton.output_indexes:
    index=o.hex()
    txo = storage_space.txos_storage.find(o, rtx=rtx)
    address=txo.address.to_text()
    lock_height = txo.lock_height
    relay_fee = txo.relay_fee
    version = txo.version
    amount = txo.value
    result['outputs'].append(({"output_id":index, "address":address, "lock_height":lock_height, "relay_fee":relay_fee, "version":version, "amount":amount}))
  return result
  
  

  
def check_sync_status(nodes, send, rtx):
  for node_index in nodes:
    node = nodes[node_index]
    if ((not "last_update" in node) or node["last_update"]+300<time()) and ((not "last_send" in node) or node["last_send"]+5<time()):
      #logger.info("\n node last_update %d was %.4f sec\n"%(("last_update" in node),   (time()-node["last_update"] if ("last_update" in node) else 0 )))
      send_tip_info(node_info = node, send=send, rtx=rtx)


def notify_all_nodes_about_tx(tx_skel, nodes, send, _except=[], mode=1):
  #TODO we should not notify about tx with low relay fee
  for node_index in nodes:
    if node_index in _except:
      continue
    node = nodes[node_index]
    send({"action":"take TBM transaction", "tx_skel": tx_skel, "mode": mode,
      "id":str(uuid4()), 'node': node["node"] })

def notify_all_nodes_about_new_tip(nodes, send, rtx, _except=[]):
  for node_index in nodes:
    node = nodes[node_index]
    if node_index in _except:
      continue
    if "height" in node:
      our_height = storage_space.blockchain.current_height(rtx=rtx)
      our_tip = storage_space.blockchain.current_tip(rtx=rtx)
      if node["height"]==our_height-1:
        serialized_header = storage_space.headers_storage.get(our_tip, rtx=rtx).serialize()
        serialized_block = storage_space.blocks_storage.get(our_tip, rtx=rtx).serialize(rtx=rtx, rich_block_format=True)
        send_headers(send, [serialized_header], [our_tip], node["node"])
        send_blocks(send, [serialized_block], [our_tip], node["node"])
    send_tip_info(node_info=node, send=send, rtx=rtx)

def send_assets(asset_type, send, serialized_assets, assets_hashes, node, _id=None):
  if not _id:
    _id=str(uuid4())
  assets, hashes = [],[]
  for i, h in enumerate(assets_hashes):
    if (node, asset_type, h) in upload_cache:
      continue
    else:
      assets.append(serialized_assets[i])
      hashes.append(h)
      upload_cache[(node, asset_type, h)] = True
  all_assets  = b"".join(assets)
  all_hashes  = b"".join(hashes)
  all_lengths = b"".join([len(a).to_bytes(2,"big") for a in assets])
  #Note hashes and lengths will be ignored by NetworkManager for headers and blocks
  if len(all_assets):
    send({"action":"take the %s"%asset_type,\
          "num": len(assets),\
          asset_type: all_assets,\
          "%s_hashes"%asset_type : all_hashes,\
          "%s_lengths"%asset_type : all_lengths,\
          "id":_id, "node": node})

def send_headers(send, headers, hashes, node, _id=None):
  send_assets("headers", send, headers, hashes, node, _id)

def send_blocks(send, blocks, hashes, node, _id=None):
  send_assets("blocks", send, blocks, hashes,  node, _id)

def send_txos(send, txos, hashes, node, _id=None):
  send_assets("txos", send, txos, hashes,  node, _id)



import shutil, os, time
from secp256k1_zkp import PrivateKey

from leer.core.lubbadubdub.ioput import IOput
from leer.core.lubbadubdub.address import Address
from leer.tests.storage.storage_for_test import TXOsStorage, wipe_test_dirs
from leer.core.storage.blocks_storage import RollBack


import random, os

def test_rollback_obj():
  for i in range(10):
    a=RollBack()
    b=RollBack()
    a.prev_state = os.urandom(random.randint(0,int(65)))
    for i in range(random.randint(0,2000)):
      a.pruned_inputs.append( [[random.randint(0,int(1e8)), b"\x03"*random.randint(0,int(100)), b"\xff"*random.randint(0,int(1000))]]*2 )
      a.pruned_inputs.append( [[0, b"", b""]]*2 )
    a.num_of_added_outputs=3
    a.num_of_added_excesses=0
    assert not len(b.deserialize_raw(a.serialize()))
    assert a.pruned_inputs ==b.pruned_inputs
    assert a.num_of_added_outputs == b.num_of_added_outputs
    assert a.num_of_added_excesses == b.num_of_added_excesses

  print("test_rollback_obj OK")







from leer.core.utils import sha256
from leer.core.storage.lookup_utils import excess_lookup, excess_lookup_by_index
from secp256k1_zkp import Point
from leer_vm import execute

def evaluate_scripts(tx, prev_block_props, excess_lookup):
  """
    Check wether all excesses in tx evaluate to True.
  """
  def find_output(commitment):
    pc = commitment.to_pedersen_commitment()
    ser = pc.serialize()
    for o in tx.outputs:
      if ser == o.serialized_apc:
        return o
    return None

  def output_lookup(commitment):
    return bool(find_output(commitment))

  burdens = []
  excesses = tx.additional_excesses + list(tx.updated_excesses.values())
  for excess in excesses:
    burden = []
    result = execute(script = excess.message,
                     prev_block_props = prev_block_props,
                     excess_lookup = excess_lookup,
                     output_lookup = output_lookup,
                     burden = burden)
    if not result:
      return False
    if len(burden):
      for comm, pubkey in burden:
        output = find_output(comm)
        pk = pubkey.to_pubkey()
        if not pk.serialize() == output.address.serialized_pubkey:
          return False
        excess_index = pk.serialize() + sha256(b"\x01\x00"+output.serialized_apc)
        burdens.append((output.serialized_index, excess_index))
  tx.burdens = burdens
  return True

def check_burdens(tx, burden_storage, excesses_storage, rtx):
  excesses = tx.additional_excesses + list(tx.updated_excesses.values())
  eis = [e.index for e in excesses]
  for i in tx.inputs:
    if burden_storage.has(i.serialized_index, rtx=rtx): 
      required_index = burden_storage.get(i.serialized_index, rtx=rtx)
      if not bool(excess_lookup_by_index(required_index, tx=tx, rtx=rtx, excesses_storage =excesses_storage)):
        return False
  return True

def generate_proof_script(output):
  script = b""
  OP_PUSHPOINT = (0x03).to_bytes(1, "big")
  OP_OUTPUTORHASH = (0x27).to_bytes(1, "big")
  OP_TRUE = (0x01).to_bytes(1, "big")
  pubkey_point = Point(output.address.pubkey)
  commitment_point = Point(output.authorized_pedersen_commitment)
  script = OP_PUSHPOINT + pubkey_point.serialize() + OP_PUSHPOINT + commitment_point.serialize() + OP_OUTPUTORHASH + OP_TRUE
  return script
  



from secp256k1_zkp import GeneratorOnCurve
from hashlib import sha256

seed = "Leer is experimental cryptocurrency implementing LubbaDubDub technology"
def generator_from_string_seed(seed):
  m=sha256()
  m.update(seed.encode())
  bytes_seed = m.digest()
  g = GeneratorOnCurve()
  g._from_seed(bytes_seed)
  g.initialise_bulletproof_generators(128)
  return g

default_generator = generator_from_string_seed(seed)
default_generator_ser=default_generator.serialize()

generators = { default_generator_ser: default_generator}



import functools
from secp256k1_zkp import PublicKey, ALL_FLAGS
import hashlib, base64

class Excess:
    '''
    Excess is public key with proof that it is public key (signature).
    '''

    def __init__(self, recoverable_signature=None, message=b'', version=0, raw=None):
        self.recoverable_signature = recoverable_signature
        self.message = message
        self.version = version
        if raw:
          self.deserialize_raw(raw)
        self.drop_cached()

    def drop_cached(self):
      self.cached_pubkey = None
      self.cached_nonrec = None
      self.serialized = None

    def calc_pubkey(self):
      unrelated = PublicKey(flags=ALL_FLAGS)
      self.cached_pubkey = PublicKey(pubkey=unrelated.ecdsa_recover(self.message, self.recoverable_signature))      
      
    @property
    def pubkey(self):
        if not self.cached_pubkey:
          self.calc_pubkey()
        return self.cached_pubkey  

    @property
    def serialized_pubkey(self):
      return self.pubkey.serialize()
        

    def from_private_key(self, privkey, message=b'', version=None):
        self.drop_cached()
        self.message = message if message else self.message
        if version:
          self.version = version
        else: 
         if self.message==b"":
           self.version = 0
         else:
           self.version = 1
        self.recoverable_signature = privkey.ecdsa_sign_recoverable(self.message)
        
        return self

    def calc_nonrec(self):
        unrelated = PublicKey(flags=ALL_FLAGS)
        self.cached_nonrec =  unrelated.ecdsa_recoverable_convert(self.recoverable_signature)

    @property
    def nonrec_signature(self):
        if not self.cached_nonrec:
          self.calc_nonrec()
        return self.cached_nonrec

    def verify(self):
        return self.pubkey.ecdsa_verify(self.message, self.nonrec_signature)

    def serialize(self):
        if self.serialized:
          return self.serialized
        if self.version==0 and len(self.message):
          raise

        unrelated = PublicKey(flags=ALL_FLAGS)
        if self.version==0:
          self.serialized = unrelated.ecdsa_recoverable_serialize_raw(self.recoverable_signature)
        elif self.version==1:
          rec_sig_serialized = unrelated.ecdsa_recoverable_serialize_raw(self.recoverable_signature)
          rec_sig_serialized = (rec_sig_serialized[0] | 128).to_bytes(1,"big") + rec_sig_serialized[1:]
          mes_serialized = len(self.message).to_bytes(2,"big")+self.message
          self.serialized = rec_sig_serialized+mes_serialized
        return self.serialized

    def deserialize_raw(self, serialized_data):
        self.drop_cached()
        consumed = b""
        if len(serialized_data)<65:
          raise Exception("Not enough bytes to encode recovery signature")
        rec_sig, serialized_data = serialized_data[:65], serialized_data[65:]
        consumed += rec_sig
        unrelated = PublicKey(flags=ALL_FLAGS)
        if rec_sig[0] & 128 ==0:
          self.version = 0
          self.message = b""
          self.recoverable_signature = unrelated.ecdsa_recoverable_deserialize_raw(rec_sig)
        if rec_sig[0] & 128 ==128:
          self.version = 1
          rec_sig = (rec_sig[0] - 128).to_bytes(1,"big") + rec_sig[1:]
          self.recoverable_signature = unrelated.ecdsa_recoverable_deserialize_raw(rec_sig)
          if len(serialized_data)<2:
            raise Exception("Not enough bytes to encode message len")
          mlen_ser, serialized_data = serialized_data[:2], serialized_data[2:]
          mlen = int.from_bytes(mlen_ser, 'big')
          if len(serialized_data)<mlen:
            raise Exception("Not enough bytes to encode message") 
          self.message, serialized_data = serialized_data[:mlen], serialized_data[mlen:]
          consumed += mlen_ser+self.message
        self.serialized = consumed
        return serialized_data      

    @property
    def index(self):
      """
        Special index which is used for building excesses merkle tree.
        Same as with tree for outputs, we want to be able to both access
        the sum of excesses as sum of points on curve and to validate tree.
        Thus index contains public key and hash of serialized excess.
      """
      m=hashlib.sha256()
      m.update(self.message)
      return self.pubkey.serialize() + m.digest()

    @classmethod
    @functools.lru_cache(maxsize=40)
    def from_serialized(cls, serialized_address):
      address = cls()
      address.deserialize_raw(serialized_address)
      return address

class Address(Excess):

    def verify(self):
        return super().verify() and (self.message==b"")


    def from_text(self, serialized):
      _address = Address()
      raw = base64.b64decode(serialized.encode())
      self.deserialize_raw(raw[:-4])
      m=hashlib.sha256()
      m.update(raw[:-4])
      checksum = m.digest()[:4]
      assert checksum==raw[-4:]
      self = _address
      return True

    def to_text(self):
      m=hashlib.sha256()
      m.update(self.serialize())
      checksum = m.digest()[:4]
      return base64.b64encode(self.serialize()+checksum).decode()

    '''
    def __init__(self, pubkey=None, signature=None, rec_sig=None, raw=False):
        self.pubkey = pubkey
        self.signature = signature
        if rec_sig:
          if raw:
            self.deserialize(rec_sig)
          else:
            self.from_recoverable_signature(rec_sig)
	
    def from_b58(self, _adr):
        pass

    def to_b58(self, _adr):
        pass

    def from_recoverable_signature(self, rec_sig):
        unrelated = PublicKey(flags=ALL_FLAGS)
        #rec_sig, rec_id = unrelated.ecdsa_recoverable_serialize(raw_rec_sig)
        self.pubkey = PublicKey(pubkey=unrelated.ecdsa_recover(b"", rec_sig))
        self.signature = unrelated.ecdsa_recoverable_convert(rec_sig)
        self.rec_sig = rec_sig

    def from_private_key(self, privkey):
        rec_sig = privkey.ecdsa_sign_recoverable(b"")
        self.from_recoverable_signature(rec_sig)
        return self

    def verify(self):
        unrelated = PublicKey(flags=ALL_FLAGS)
        return self.pubkey.ecdsa_verify(b"", self.signature)

    def serialize(self):
        unrelated = PublicKey(flags=ALL_FLAGS)
        return unrelated.ecdsa_recoverable_serialize_raw(self.rec_sig)

    def deserialize(self, ser_rec_sig):
        unrelated = PublicKey(flags=ALL_FLAGS)
        self.from_recoverable_signature(unrelated.ecdsa_recoverable_deserialize_raw(ser_rec_sig))
    '''


def excess_from_private_key(private_key, apc):
  e= Excess(message=apc)
  e.from_private_key(private_key)
  return e

def address_from_private_key(private_key):
  a= Address(message=b"")
  a.from_private_key(private_key)
  return a



#class Excess(Address, PublicKey):
#    def from_address(self, adr):
#        PublicKey.__init__(self, pubkey=adr.pubkey.public_key, raw=False)





secp256k1_order = 0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141

def sum_offset(offset1, offset2):
  return (offset1 + offset2) % secp256k1_order

def subtract_offset(offset1, offset2):
  return (offset1 + secp256k1_order - offset2) % secp256k1_order



from collections import OrderedDict
import os, lmdb, math


def _(x):
  return (x).to_bytes(5,'big')

class MMR:
  """
    Storage with merkle tree structure above it.

    We add obj and also obj index (sometimes term `hash` is used instead of `index`).
    Tree is built from indexes via summ which should be redefined in subclass.
                     
                       root = i4_1=sum(i3_1,i3_2)
                     /                           \
             i3_1=sum(i2_1,i2_2)             i3_2=i2_3
             /                 \                     |
     i2_1=sum(i1,i2)    i2_2=sum(i3,i4)      i2_3=sum(i5,i6)
       /        \          /        \         /         \
      i1        i2        i3        i4        i5        i6 
      |         |         |         |         |         |         
    obj1      obj2      obj3      obj4      obj5      obj6      

  Boundary case: tree with zero leafs has root equal to `'\\x00'*default_index_size`

  Inner structure contains 4 lmdb databases: leaf_db, node_db, order_db, reverse_order_db.
  order_db: 		sequence_num -> index
  reverse_order_db: 	index -> sequence_num
  leaf_db: 		index -> object
  node_db: 		(level,sequence_num) -> tree_node_index

  3 remove-like operations are supported:
    remove(n)  - removes n elements from the end of the list. Cannot be undone.
                 This operation is used to delete information from orphan blocks.
                 This operation is available for all trees
    clear(index) - delete obj with `index`, however keep its place (any other elements occupies
                  after this operation the same places). Hash of place becomes
                  equal None: no matter how summ of indexes is redefined in subclasses
                  `None`+`smth`=`smth`, `None`+`None`=`None`. This operation is used for
                  unspent outputs tree.
    discard(index) - delete obj, but keeps its index. This operation is used for txos tree.

  Special value is leaf db with index `b'state'. It is used to store identifier of current MMR state on disc.
    `set_state` and `get_state` write and read this value
  """
  def __init__(self, name, dir_path, env, wtx, waterline_depth=16, default_index_size=65, discard_only=False, clear_only=False, save_pruned=True):
    self.index={}
    self.dir_path = dir_path
    self.name = bytes(name.encode("utf-8"))
    self.default_index_size = default_index_size
    self.discard_only = discard_only
    self.clear_only = clear_only
    self.save_pruned = save_pruned # Do not really delete data from db for debug purposes


    self.env = env
    self.leaf_db = self.env.open_db(self.name+b'leaf_db', txn=wtx)
    self.node_db = self.env.open_db(self.name+b'node_db', txn=wtx)
    self.order_db = self.env.open_db(self.name+b'order_db', txn=wtx)
    self.reverse_order_db = self.env.open_db(self.name+b'reverse_order_db', txn=wtx, dupsort=True)
    if self.save_pruned:
        self.pruned_db = self.env.open_db(self.name+b'pruned_db', txn=wtx)
        self.pruned_ro_db = self.env.open_db(self.name+b'pruned_ro_db', txn=wtx, dupsort=True)

  def _get_node(self, level, sequence_num, rtx):
    if level==0:
          return rtx.get( _(sequence_num), db=self.order_db)
    else:
        return rtx.get(_(level-1)+_(sequence_num), db=self.node_db)

  def get_by_hash(self, _hash, rtx):
    return rtx.get(_hash, db=self.leaf_db)

  def fuzzy_search(self, _hash, rtx):
    cursor = rtx.cursor(db=self.leaf_db)
    is_set = cursor.set_range(_hash)
    if not is_set:
      return None
    return cursor.item()

  def get_pruned_by_hash(self, _hash, rtx):
    if not self.save_pruned:
      raise
    return rtx.get(_hash, db=self.pruned_db)

  def find_by_hash(self, _hash, rtx):
    '''
      In contrast with get_by_hash, find_by_hash tries to find result both in existing and pruned(if saved) dbs.
    '''
    r1 = rtx.get(_hash, db=self.leaf_db) 
    if not r1 and self.save_pruned:
     r1 = rtx.get(_hash, db=self.pruned_db)
    return r1

  def _set_node(self, level, sequence_num, value, wtx):
    wtx.put(_(level-1)+_(sequence_num), value, db=self.node_db)

  def _del_node(self, level, sequence_num, wtx):
    wtx.delete(_(level-1)+_(sequence_num), db=self.node_db)

  def num_of_elements(self, rtx):
    return rtx.stat(db=self.order_db)['entries']

  def _get_max_level(self, rtx):
    try:
      return math.ceil(math.log(self.num_of_elements(rtx=rtx),2))
    except ValueError:
      #if num_of_elements ==0, math.log raise ValueError
      raise ValueError("getting root level of empty tree")

  def _update_path(self, level, sequence_num, wtx):
    """
      This function should be called when node on level `level` with sequence_num `sequence_num` changes.
      It updates all nodes above
    """
    try:
      mxlvl = self._get_max_level(rtx=wtx)
    except ValueError:
      return
    if level == mxlvl and sequence_num==0:
      #already at the root, nothing to update
      pass
    else:
      neighbour_index = (sequence_num//2)*2 if (sequence_num%2) else (sequence_num//2)*2+1
      neighbour = self._get_node(level, neighbour_index, rtx=wtx)
      we = self._get_node(level, sequence_num, rtx=wtx)
      if neighbour and we:
        if sequence_num%2:
          new_value = self.sum(neighbour, we)
        else:
          new_value = self.sum(we, neighbour)
      else:
        new_value = we if we else neighbour# just copy value up
      self._set_node(level+1, sequence_num//2, new_value, wtx=wtx)
      self._update_path(level+1, sequence_num//2, wtx=wtx)

  def append(self, wtx, obj_index=None, obj=None):
      num = wtx.stat(db=self.order_db)['entries']
      #TODO actually we probably may use append=True flag for order_db
      wtx.put( _(num), bytes(obj_index), db=self.order_db)
      wtx.put( bytes(obj_index), _(num), db=self.reverse_order_db)
      wtx.put(bytes(obj_index), bytes(obj), db=self.leaf_db)
      self._update_path(0, num, wtx=wtx)
      return _(num)

  def has_index(self, rtx, obj_index):
      index = rtx.get( bytes(obj_index), db=self.reverse_order_db)
      return bool(index)

  def append_unique(self, wtx, obj_index=None, obj=None):
      index = wtx.get( bytes(obj_index), db=self.reverse_order_db)
      if index:
        raise KeyError("Not unique")
      self.append(wtx=wtx, obj_index=obj_index, obj=obj)

  def update_index_by_num(self, wtx, num_index_ser, obj_index, obj):
      old_index = wtx.get( num_index_ser, db=self.order_db)
      wtx.put( num_index_ser, bytes(obj_index), db=self.order_db)
      wtx.delete( bytes(old_index), num_index_ser, db=self.reverse_order_db)
      wtx.put( bytes(obj_index), num_index_ser, db=self.reverse_order_db)
      if not wtx.get(bytes(old_index), db=self.reverse_order_db):
        #no leafs with the same index      
        old_obj = wtx.pop(bytes(old_index), db=self.leaf_db)
      else:
        old_obj = wtx.get(bytes(old_index), db=self.leaf_db)
      wtx.put(bytes(obj_index), bytes(obj) , db=self.leaf_db)
      self._update_path(0, int.from_bytes(num_index_ser,"big"), wtx=wtx)
      return old_index, old_obj

  def update_index_by_num_unique(self, wtx, num_index_ser, obj_index, obj):
      nindex = wtx.get( bytes(obj_index), db=self.reverse_order_db)
      if nindex:
        raise KeyError("Not unique")
      return self.update_index_by_num(wtx, num_index_ser, obj_index, obj)

      
    

  def remove(self, num, wtx, set_of_indexes=None):
    """
     Remove `num_of_elements` from right end.
     If `set_of_elements` specified check before remove that all elements are in set.
    """
    start_n = self.num_of_elements(rtx=wtx)-1
    mxlvl = self._get_max_level(rtx=wtx) 
    cache = {}
    if set_of_indexes:
        for el_n in range(start_n, start_n-num, -1):
          el=wtx.get(_(el_n),db=self.order_db)
          cache[el_n] = el
          if not el in set_of_indexes:
            raise #TODO
    removed_objects = []
    el_n = start_n
    for el_n in range(start_n, start_n-num, -1):
        if not el_n in cache:
          el=wtx.get(_(el_n),db=self.order_db)
          cache[el_n] = el
        wtx.delete( _(el_n), db=self.order_db)
        wtx.delete( cache[el_n], _(el_n), db=self.reverse_order_db)
        if not wtx.get(cache[el_n], db=self.reverse_order_db):
          obj = wtx.pop( cache[el_n], db=self.leaf_db)
        else:
          obj = wtx.get( cache[el_n], db=self.leaf_db)
        removed_objects.append(obj)
        #TODO we can make this much faster for bulk deletion by not updating
        # each path separately
        for l in range(1,mxlvl+1):
          self._del_node(l, el_n//(2**l), wtx=wtx) 
    self._update_path(0, el_n-1, wtx=wtx)
    return removed_objects

  def discard(self, _index, wtx):
    '''
      Discard leaf, that means delete obj by index(for saving space), but keep its index.
      Returns object which can be used to revert discarding.
    '''
    #TODO: check wether both children are discarded (thus we can discard childs and keep only parent index)
    if self.clear_only:
      raise
    num = int.from_bytes(wtx.get( bytes(_index), db=self.reverse_order_db), 'big')
    #wtx.put( _(num), b"", db=self.order_db)
    obj = wtx.pop( bytes(_index), db=self.leaf_db)
    if self.save_pruned:
        wtx.put(_(num), bytes(_index), db=self.pruned_db) 
        wtx.put(bytes(_index), obj, db=self.pruned_db)       
    self._update_path(0, num, wtx=wtx)
    return [num, _index, obj]

  def revert_discarding(self, prune_obj, wtx):
    '''
      Revert discarding by object which is returned by discard
    '''
    num, obj_index, obj = prune_obj
    wtx.put( _(num), bytes(obj_index), db=self.order_db)
    wtx.put( bytes(obj_index), _(num), db=self.reverse_order_db)
    wtx.put(bytes(obj_index), bytes(obj), db=self.leaf_db) 
    if self.save_pruned:
        wtx.pop(_(num), db=self.pruned_db) 
        wtx.pop(bytes(obj_index), db=self.pruned_db)        
    self._update_path(0, num, wtx=wtx)  

  def clear(self, _index, wtx):
    '''
      Clear leaf, that means delete information (index and object), 
      but keep place. Returns prune object by which it can be easily unpruned
    '''
    if self.discard_only:
      raise
    num = int.from_bytes(wtx.pop( bytes(_index), db=self.reverse_order_db), 'big')
    wtx.put( _(num), b"", db=self.order_db)
    obj = wtx.pop( bytes(_index), db=self.leaf_db)
    if self.save_pruned:
        wtx.put(_(num), bytes(_index), db=self.pruned_db) 
        wtx.put(bytes(_index), obj, db=self.pruned_db) 
        wtx.put(bytes(_index), _(num), db=self.pruned_ro_db)           
    self._update_path(0, num, wtx=wtx)
    return [num, _index, obj]

  def revert_clearing(self, prune_obj, wtx):
    '''
      revert_clearing takes object which clear function returns and reverts clear operation.
    '''
    num, obj_index, obj = prune_obj
    wtx.put( _(num), bytes(obj_index), db=self.order_db)
    wtx.put( bytes(obj_index), _(num), db=self.reverse_order_db)
    wtx.put(bytes(obj_index), bytes(obj), db=self.leaf_db)  
    if self.save_pruned:
        wtx.pop(_(num), db=self.pruned_db) 
        wtx.pop(bytes(obj_index), db=self.pruned_db) 
        wtx.pop(bytes(obj_index), db=self.pruned_ro_db)   
    self._update_path(0, num, wtx=wtx) 

  def sum(self, x1, x2):
    """
      Should be redefined in subclasses
    """
    pass


  def get_root(self, rtx):
    try:
      return self._get_node(self._get_max_level(rtx=rtx), 0, rtx=rtx)
    except ValueError:
      #empty tree (TODO custom exception)
      return b"\x00"*self.default_index_size

  def set_state(self, state, wtx):
    wtx.put(b'state', state, db=self.leaf_db)


  def get_state(self, rtx):
    return rtx.get(b'state', db=self.leaf_db)





from time import time, sleep

def notification_center_launcher(syncer, config):
  '''
    Notification center is generally memcache. It has no inner logic and thus is synchronous
  '''
  message_queue = syncer.queues['Notifications']
  keyvalue = {}
  while True:
    sleep(0.001)
    while not message_queue.empty():
      message = message_queue.get()
      if not 'action' in message:
        continue
      if not message['action'] in ['set', 'get']:
        continue
      if not 'time' in message:
        message['time'] = time()
      if message['action']=="set":
        result = None
        try:
          keyvalue[message['key']]={'value':message['value'], 'time':message['time']}
          result = True
        except KeyError as e:
          result = False
          error = e
        if ('id' in message) and ('sender' in message):
          if result:
            syncer.queues[message['sender']].put({'id':message['id'], 'result':'set'})
          else:
            syncer.queues[message['sender']].put({'id':message['id'], 'result':'error', 'error':str(e)})
      if message['action']=="get":
        error=None
        if (not 'id' in message) or (not 'sender' in message):
          continue
        try:
          result = keyvalue[message['key']]
        except KeyError as e:
          result = None
          error = e
        if result:
            syncer.queues[message['sender']].put({'id':message['id'], 'result':result})
        else:
            syncer.queues[message['sender']].put({'id':message['id'], 'result':'error', 'error':str(error)})
      if message['action']=="stop":
        return



import struct
import hashlib

from secp256k1_zkp import PrivateKey, PedersenCommitment, RangeProof, BulletProof

from leer.core.lubbadubdub.constants import default_generator, default_generator_ser, generators
from leer.core.lubbadubdub.address import Address
from leer.core.lubbadubdub.utils import encrypt, decrypt
from leer.core.storage.verification_cache import verification_cache

def is_sorted(lst, key=lambda x: x):
    for i, el in enumerate(lst[1:]):
        if key(el) < key(lst[i]): # i is the index of the previous element
            return False
    return True


class IOput:
  """ Basic class which represent transaction input or output"""

  # Pedersen commitment should be unique, it is id of IOput
  def __init__(self, json_object=None, binary_object=None):
    """
    Initialize empty ioput

    Parameters
    ----------
    [optional] binary_object : bytes
        serialized output.
    """
    #serializable data
    self.version = 2
    self.block_version = 1
    self.lock_height = 0
    self.authorized_pedersen_commitment = None
    self.address = None
    self.rangeproof = None
    self.encrypted_message = None
    self.generator = None
    self.relay_fee = None
    #context data (we serialize it for storing, but not for sending to network)
    self.address_excess_num_index = None #5 bytes long bytes-string
    #inner data
    self.unauthorized_pedersen_commitment = None
    self.value = None
    self.blinding_key = None
    self.serialized = None
    self._serialized_apc = None

    if json_object:
      self.from_json(json_object)
    if binary_object:
      self.deserialize(binary_object)

  def from_json(self, json_object):
    pass  #TODO

  def to_json(self):
    pass #TODO

  def serialize(self):
    """ Returns binary representation of output"""
    if self.serialized:
      return self.serialized
    ret = self.signed_part()

    ser_range_proof = self.rangeproof.proof
    ret += struct.pack("> H", len(ser_range_proof))
    ret += ser_range_proof

    self.serialized = ret
    return ret

  def serialize_with_context(self):
    ret = self.serialize()
    ret += self.address_excess_num_index
    return ret

  @property
  def hash(self):
    seed = self.serialize()
    m=hashlib.sha256()
    m.update(seed)
    return m.digest()

  @property
  def index_len(self):
    #len apc + len hash
    return 33+32

  @property
  def serialized_apc(self):
    if not self._serialized_apc:
      self._serialized_apc = self.authorized_pedersen_commitment.serialize()
    return self._serialized_apc

  @property
  def serialized_index(self):
    """
      Returns binary representation of output index.
      For transaction to be spent, it is necessary that all inputs are known.
      Thus it is possible to include only index for inputs.
      Since all outputs should have unique authorized pedersen commitments(APC),
      it seems expedient to use APC as index. However it may cause issues with
      transaction which spends from memory pool. Indeed, while confirmed 
      transactions strictly have unique APC, different nodes may have transaction
      with different outputs but with the same APC for those outputs (_double spend_) in memory

      Then, if another transaction which spends _double spending_ will be added
      to memory pool, different nodes may have different opinions on its validity.
      It is possible to use such dis-consensus as an attack to disrupt connectivity.

      Another option is hash of outputs. However, hash as index hinders some
      basic mimblewimble checks: we can't check that sum of inputs and outputs is
      zero without access to ledger. In our terms, all checks become context
      dependent.
      Thus we use concantenation of APC and hash. It's slightly bigger (although nothing
      in comparisson with rangeproof) but allows both basic checks and unambiguous
      indexing.
    """
    return self.serialized_apc+self.hash

  @property
  def commitment_index(self):
    """
      Special index which is used for building commitment merkle tree
    """
    m=hashlib.sha256()
    m.update(self.serialized_apc)
    return self.serialized_apc + m.digest()

  @property
  def is_coinbase(self):
    return self.version==0

  def deserialize(self, serialized_output):
    """ Decode output from serialized representation. """
    self.serialized = None
    self._serialized_apc = None
    self.deserialize_raw(serialized_output)

  def deserialize_raw(self, serialized_output):  
    #TODO part1, part2, part3 should be substitued with construction `something, serialized = serialized[:x], serialized[x:]`
    # as it is done in other modules
    """ Decode output from serialized representation. Return residue of data after serialization"""
    self.serialized = None
    self._serialized_apc = None
    consumed = b""

    if len(serialized_output)<145:
        raise Exception("Serialized output doesn't contain enough bytes for constant length parameters")

    part1, part2 = serialized_output[:82], serialized_output[82:]
    (self.version, self.block_version, self.lock_height,
      self.generator, self.relay_fee, self.apc) = struct.unpack("> H H L 33s Q 33s", part1) 
    consumed+=part1

    if self.generator in generators:
      self.authorized_pedersen_commitment = PedersenCommitment(commitment=self.apc, raw=True, value_generator = generators[self.generator])
    else:
      raise NotImplemented

    self.address = Address()
    
    _part2 = self.address.deserialize_raw(part2)
    consumed += part2[:len(part2)-len(_part2)]
    part2 = _part2
    
    if len(part2)<2:
        raise Exception("Serialized output doesn't contain enough bytes for encrypted message length")
    (encrypted_message_len,) = struct.unpack("> H", part2[:2]) 
    if len(part2)<2+encrypted_message_len:
        raise Exception("Serialized output doesn't contain enough bytes for encrypted message")
    self.encrypted_message = part2[2:2+encrypted_message_len]
    consumed += part2[:2+encrypted_message_len]


    part3=part2[2+encrypted_message_len:]
    (range_proof_len,) = struct.unpack("> H", part3[:2])
    if len(part3)<2+range_proof_len: 
        raise Exception("Serialized output doesn't contain enough bytes for rangeproof")

    self._calc_unauthorized_pedersen()  
    if self.version in [0,1]:
      self.rangeproof = RangeProof(proof=part3[2:2+range_proof_len], 
          pedersen_commitment=self.unauthorized_pedersen_commitment, 
          additional_data = self.signed_part())
    elif self.version == 2:
      self.rangeproof = BulletProof(proof=part3[2:2+range_proof_len], 
          pedersen_commitment=self.unauthorized_pedersen_commitment, 
          additional_data = self.signed_part())      

    consumed += part3[:2+range_proof_len]

    info=self.info()
    if info['min_value']==info['max_value']:
      self.value=info['min_value']
    self.serialized = consumed
    return part3[2+range_proof_len:]


  def deserialize_with_context(self, serialized_output):
    residue = self.deserialize_raw(serialized_output)
    self.address_excess_num_index, residue = residue[:5], residue[5:]
    return residue

  def detect_value(self, inputs_info):
    try:
          privkey = inputs_info['priv_by_pub'][self.address.serialized_pubkey]
          nonce = self.apc
          decrypted_message = decrypt(privkey, nonce, self.encrypted_message)
          raw_blinding_key, self.value = struct.unpack( "> 32s Q", decrypted_message)
          self.blinding_key=PrivateKey(raw_blinding_key, raw=True)
          if not self._calc_pedersen_wos()==self.unpc:
            self.blinding_key, self.value = None,None
            raise Exception("Incorrect blinding key and value")
    except KeyError as e:
         raise e #Wrong inputs_info
    except Exception as e:
         #TODO definetely some logic should be added here to notify about missed info
         pass  
    return bool(self.value)
       

  def signed_part(self):
    """
      Flurbo output has more abundant structure than classical mw output.
      To be sure that all elements of structure were relayed unchanged, all
      elements are signed in rangeproof. This function returns serialized
      representation elements which should be signed.
    """
    ret = b''
    ret += struct.pack("> H H L 33s Q 33s",
      self.version, self.block_version, self.lock_height,
      self.generator, self.relay_fee,
      self.serialized_apc)
    ret += self.address.serialize()
    ret += struct.pack("> H", len(self.encrypted_message))
    ret += self.encrypted_message
    return ret


  #Wallet functionality
  def fill(self, address, value, relay_fee = 0, blinding_key=None, generator=default_generator_ser, coinbase=False, lock_height = 0):
    """
    Fill basic params of ouput.

    Fill params, but doesn't generate commitments and rangeproofs.
    It should be done separately by generate() function

    Parameters
    ----------
    address : Address
        Address of output
    value : int
        Value of output in minimal indivisible units
    [optional] relay_fee : int
        Default:0. Relay fee of output in minimal indivisible units
    [optional] blinding_key : PrivateKey
        Default: random. Usually wallet should not store this key: it will
        be encrypted it the output with private key of address
    [optional] generator : GeneratorOnCurve
        Default: default_generator. Asset transactions will use another generators
    [optional] coinbase : bool
        Default: False. Coinbase outputs have version 0 (while common version is 1)
    [optional] lock_height : integer
        Default: 0. Minimal height at which output can be spent.
    """
    self.serialized = None
    self._serialized_apc = None
    self.address = address
    self.generator = generator
    self.value = value
    if blinding_key is None:
      blinding_key = PrivateKey() #we do not store this key in the wallet
    self.blinding_key = blinding_key
    self.relay_fee = relay_fee
    if coinbase:
      self.version=0
    self.lock_height = lock_height


  def _calc_pedersen_wos(self):
    """
      Calc unauthorized pedersen commitment from generators, blinding key
      and value.
      Note, it differs from _calc_unauthorized_pedersen (which calc UPC
      from APC and address).
    """
    assert(self.generator and self.address and self.blinding_key and isinstance(self.value, int)) #self.value can be 0
    if self.generator in generators:
      unpc = PedersenCommitment(value_generator = generators[self.generator])
    else:
      raise NotImplemented #TODO
    unpc.create(self.value, self.blinding_key.private_key)
    return unpc

  # rename? 
  def _calc_pedersen(self):
    self.unauthorized_pedersen_commitment = self._calc_pedersen_wos()
    

  def _calc_authorized_pedersen(self):
    """
      Calc authorized pedersen commitment from UPC and address.
    """
    assert self.unauthorized_pedersen_commitment
    assert self.generator
    self._serialized_apc = None
    self.authorized_pedersen_commitment = \
      (self.unauthorized_pedersen_commitment.to_public_key() + self.address.pubkey).to_pedersen_commitment(
      value_generator = generators[self.generator])

  def _calc_unauthorized_pedersen(self):
    """
      Calc unauthorized pedersen commitment from APC and address.
    """
    assert self.authorized_pedersen_commitment
    assert self.generator
    if not self.generator in generators:
      raise NotImplemented
    self.unauthorized_pedersen_commitment = \
      (self.authorized_pedersen_commitment.to_public_key() - self.address.pubkey).to_pedersen_commitment(
      value_generator = generators[self.generator])

  #TODO default exp should be more wise
  def generate(self, min_value=0, nonce=None, exp=0, concealed_bits=64):
    """
    Generate output

    Calc all necessery params like APC, rangeproofs and so on. After generation
    ouput is ready for serialization. Params listed below control what should be
    concealed by proof.
    Note, if version = 2 is set, bullerproofs with min_value equal to 0, 
    concealed_bits = 64 are used. All optional params are neglected

    Parameters
    ----------
    [optional] min_value : int
        Default: 0. Constructs a proof where the verifer can tell the minimum
                   value is at least the specified amount.
    [optional] exp : int
        Default: 0. Base-10 exponent. Number of digits which will be made public:
                   Allowed range is -1 to 18
                   0 corresponds to all digits (with respect to concealed bits) are private,
                   1 corrsponds to smallest digit are public etc
                   -1 is special value to make all digits public.
    [optional] nonce : 32bytes
        Default: random. 32-byte secret nonce used to initialize the proof.
    [optional] concealed_bits : int
        Default:64. Number of bits of the value to keep private. 
                    (0 = auto/minimal, 64). For instance minimal number of bits
                    to conceal value 100 is 7. For proof with 7 bits observer 
                    can check that value is between 0 and 127. If minimal number
                    of bits will be used chainwide observer can guess with a certain 
                    degree of confidence that value is between 64 and 127.
                    
    """    
    self.serialized = None
    self._serialized_apc = None
    self._calc_pedersen()
    self._calc_authorized_pedersen()

    apc = self.serialized_apc 
    plaintext = struct.pack( "> 32s Q", self.blinding_key.private_key, self.value)
    self.encrypted_message = encrypt(self.address.pubkey, apc, plaintext);

    additional_data = self.signed_part()
    if self.version==0:
      self.rangeproof = RangeProof(pedersen_commitment=self.unauthorized_pedersen_commitment, 
                                   additional_data = additional_data)
      res = self.rangeproof._sign(exp=-1, concealed_bits=0,
                                  nonce=nonce)
    elif self.version==1:
      self.rangeproof = RangeProof(pedersen_commitment=self.unauthorized_pedersen_commitment, 
                                   additional_data = additional_data)
      self.rangeproof._sign(min_value=min_value, nonce=nonce,
                        exp=exp, concealed_bits=concealed_bits)
    elif self.version==2:
      self.rangeproof = BulletProof(pedersen_commitment=self.unauthorized_pedersen_commitment, 
                                   additional_data = additional_data)
      self.rangeproof._sign(concealed_bits=64)
    
  def set_verified_and_correct(self):
    verification_cache[self.serialize()] = True

  #rename to validate?
  def verify(self):
    """
     Verify IOput.
     
     Requirements for valid ouput:
     0) version is known
     1) address is valid
     2) generator is known
     3) range_proof is valid and signs (version, address, asset_type, encrypted_message, relay_fee)
    """
    try:
      return verification_cache[self.serialize()]
    except KeyError:
      pass
    result = True

    try:
      assert self.address.verify(), "Bad address"
      assert self.generator in generators, "Bad generator"
    except AssertionError as e:
      result = False

    if result:
      if self.version==1 or self.version==0:
        try:
          assert self.rangeproof.verify(), "Bad rangeproof"
        except AssertionError as e:
          result = False    
      elif self.version==2:
        try:
          assert self.rangeproof.verify(concealed_bits = 64), "Bad bulletproof"
        except AssertionError as e:
          result = False  
      else:
        result = False

    verification_cache[self.serialize()] = result
    return result

  def info(self):
    """
      Returns dictionary with params extracted from proof:
      'exp', 'mantissa' (the same as concealed bits), 
      'min_value', 'max_value'
    """
    if self.version in [0,1]:
      #Rangeproof
      a,b,c,d =  self.rangeproof.info()
      return {'exp':a, 'mantissa':b, 'min_value':c, 'max_value':d}
    if self.version == 2:
      #Bulletproof, no info here
      assert self.verify() 
      return {'exp':0, 'mantissa':0, 'min_value':0, 'max_value': 18446744073709551615} # 18446744073709551615==2**64-1

  def __str__(self):
    s=""
    s+="Output[ coinbase: %s, Pedersen: 0x%s, Pubkey: 0x%s, RangeProof: %s, Value: %s, Fee: %s, Blinding key: %s...]" %(
      "+" if self.version==0 else "-", 
      self.serialized_apc[:8].hex(),
      self.address.pubkey.serialize()[:10].hex(),
      "+" if self.rangeproof else "-", 
      str(self.value) if self.value else 'unknown',
      str(self.relay_fee), 
      ('0x'+self.blinding_key.serialize()[:8]) if self.blinding_key else 'unknown', )
    return s



from collections import OrderedDict, namedtuple
from functools import partial
import struct, os

from secp256k1_zkp import PrivateKey, PublicKey, PedersenCommitment, RangeProof, default_blinding_generator, Point

from leer.core.lubbadubdub.offset_utils import sum_offset
from leer.core.lubbadubdub.constants import default_generator, default_generator_ser, generators
from leer.core.lubbadubdub.address import Address, Excess, excess_from_private_key
from leer.core.lubbadubdub.ioput import IOput
from leer.core.storage.txos_storage import TXOsStorage


from leer.core.parameters.constants import coinbase_maturity, output_creation_fee
from leer.core.primitives.transaction_skeleton import TransactionSkeleton
from leer.core.storage.verification_cache import verification_cache
from leer.core.lubbadubdub.script import evaluate_scripts, check_burdens, generate_proof_script
from leer.core.storage.lookup_utils import excess_lookup

def is_sorted(lst, key=lambda x: x):
    for i, el in enumerate(lst[1:]):
        if key(el) < key(lst[i]): # i is the index of the previous element
            return False
    return True

pseudoinput = namedtuple('pseudoinput',('serialized_index', 'serialized_apc'))


class Transaction:

  def __init__(self, txos_storage, excesses_storage,  raw_tx=None):
    #serializable data 
    self.inputs = []
    self.updated_excesses = {} # after spending inputs their addresses excesses should be updated to become additional excesses
    self.outputs = []
    self.additional_excesses = []
    self.mixer_offset = 0
    #inner data
    self._destinations = [] #destinations are unprepared outputs
    self.coinbase = None
    self.new_outputs_fee = None
    self.txid = None
    self.serialized = None

    self.txos_storage = txos_storage 
    self.excesses_storage = excesses_storage
    
    if raw_tx:
      self.deserialize(raw_tx)

  def push_input(self, _input):
    self.inputs.append(_input)

  def add_destination(self, destination):
    '''
      Destination is address, value, [optional] need_proof
    '''
    if len(destination)==2:
      destination = (destination[0], destination[1], True)
    self._destinations.append(destination)


  #TODO coinbase should be @property

  def add_coinbase(self, coinbase_output):
    self.serialized = None
    self.coinbase = coinbase_output

  def compose_block_transaction(self, rtx, combined_transaction=None):
    self.serialized = None
    if not self.coinbase:
      raise Exception("coinbase output is required")
    cb = self.coinbase
    self.__init__(txos_storage = self.txos_storage, excesses_storage = self.excesses_storage) #reset self
    self.outputs = [cb]
    offset_pk = PrivateKey()
    self.mixer_offset = int.from_bytes(offset_pk.private_key, "big")
    self.additional_excesses = [excess_from_private_key(cb.blinding_key+offset_pk, b"\x01\x00"+cb.serialized_index[:33])]
    if combined_transaction:
      new_tx = self.merge(combined_transaction)
      self.inputs = new_tx.inputs
      self.outputs = new_tx.outputs
      self.additional_excesses = new_tx.additional_excesses
      self.updated_excesses = new_tx.updated_excesses
    self.verify(rtx=rtx)

  def blindly_generate(self, change_address, input_data, relay_fee_per_kb=0):
    self.serialized = None
    if self.coinbase:
      raise Exception("generate() can be used only for common transaction, to create block transaction as miner use compose_block_transaction")
    if not len(input_data):
      raise Exception("Tx should have at least one input")
    if not len(self._destinations):
      raise Exception("Tx should have at least one destination")
    in_value = sum([ioput[1] for ioput in input_data]) 
    out_value = sum([destination[1] for destination in self._destinations])
    relay_fee = self.calc_relay_fee(relay_fee_per_kb=relay_fee_per_kb)
    # +1 for destination is for change address
    self.fee = relay_fee + self.calc_new_outputs_fee(len(input_data), len(self._destinations)+1)
    remainder = in_value - out_value - self.fee
    if remainder<0:
      raise Exception("Not enough money in inputs to cover outputs")
    # TODO We need logic here to cover too low remainders (less than new output fee)
    self._destinations.append((change_address, remainder, True))
    privkey_sum=0
    out_blinding_key_sum = None
    need_proofs = []
    excesses_key_sum = None
    for out_index in range(len(self._destinations)-1):
      address, value, need_proof = self._destinations[out_index]
      output = IOput()
      output.fill(address, value, generator = default_generator_ser)
      self.outputs.append( output )
      out_blinding_key_sum = (out_blinding_key_sum + output.blinding_key) if out_blinding_key_sum else output.blinding_key
      if need_proof:
        need_proofs.append((output, PrivateKey())) #excesses will be generated after output generation
    offset_pk = PrivateKey()
    self.mixer_offset = int.from_bytes(offset_pk.private_key, "big")
    # privkey for the last one output isn't arbitrary
    address, value, need_proof = self._destinations[-1]
    if need_proof:
      need_proofs.append((output, PrivateKey())) #excesses will be generated after output generation
    in_blinding_key_sum = None
    burdens_to_be_covered = []
    for i, v, priv_key, blinding_key, ser_apc in input_data:
      in_blinding_key_sum = (in_blinding_key_sum + blinding_key) if in_blinding_key_sum else blinding_key
      in_blinding_key_sum += priv_key
      self.updated_excesses[i]=excess_from_private_key(priv_key, b"\x01\x00"+ser_apc)
      self.inputs.append(pseudoinput(i, ser_apc)) #For tx serialization and sorts
    if len(need_proofs):
      excesses_key_sum = need_proofs[0][1]
      for i in need_proofs[1:]:
        excesses_key_sum+=i[1]
    output = IOput()
    last_blinding_key = in_blinding_key_sum-out_blinding_key_sum-offset_pk
    if excesses_key_sum:
      last_blinding_key += excesses_key_sum
    output.fill(address, value, blinding_key = last_blinding_key,
      relay_fee=relay_fee, generator = default_generator_ser) #TODO relay fee should be distributed uniformly, privacy leak
    self.outputs.append(output)
    [output.generate() for output in self.outputs]
    for ae in need_proofs:
      script = generate_proof_script(ae[0])
      e = excess_from_private_key(ae[1], script)
      self.additional_excesses.append(e)
    self.sort_lists()
    

  def calc_relay_fee(self, relay_fee_per_kb):
    inputs_num, outputs_num, excesses_num = len(self.inputs), len(self._destinations)+1, len(self.additional_excesses)
    input_size = 67+2
    output_size = 5366+2
    excess_size = 65+2
    mixer_offset_size = 32
    estimated_size = 6+inputs_num*input_size + outputs_num*output_size + excesses_num*excess_size + mixer_offset_size
    return int((estimated_size/1000.)*relay_fee_per_kb)


  def calc_new_outputs_fee(self, inputs_num=None, outputs_num=None):
    return ( (outputs_num if outputs_num else len(self.outputs)) -
             (inputs_num if inputs_num else len(self.inputs)) - 
             (1 if self.coinbase else 0))*output_creation_fee

  def sort_lists(self):
      self.inputs = sorted(self.inputs, key= lambda _input: _input.serialized_apc)
      self.outputs = sorted(self.outputs, key= lambda _output: _output.serialized_apc)
      self.additional_excesses = sorted(self.additional_excesses, key = lambda e: e.index)

  def serialize(self):
    if self.serialized:
      return self.serialized
    ret=b""
    ret += struct.pack("> H", len(self.inputs)) 
    for _input in self.inputs:
        s_i = _input.serialized_index
        ret += struct.pack("> H", len(s_i))
        ret += s_i

    ret += struct.pack("> H", len(self.outputs)) 
    for _output in self.outputs:
        s_o = _output.serialize()
        ret += struct.pack("> H", len(s_o))
        ret += _output.serialize()

    ret += struct.pack("> H", len(self.additional_excesses)) 
    for _excess in self.additional_excesses:
        s_e = _excess.serialize()
        ret += struct.pack("> H", len(s_e))
        ret += s_e
    for _input in self.inputs:
        s_ue = self.updated_excesses[_input.serialized_index].serialize()
        ret += struct.pack("> H", len(s_ue))
        ret += s_ue
    ret += self.mixer_offset.to_bytes(32, "big")
    self.serialized = ret
    return ret

  def deserialize(self, serialized_tx, rtx, skip_verification=False):
    self.serialized = None
    if len(serialized_tx)<2:
        raise Exception("Serialized transaction doesn't contain enough bytes for inputs array length")
    inputs_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
    (inputs_len,) = struct.unpack("> H", inputs_len_buffer) 
    for _input_index in range(inputs_len):
        if len(serialized_tx)<2:
          raise Exception("Serialized transaction doesn't contain enough bytes for input %s length"%_input_index)
        input_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
        (input_len,) = struct.unpack("> H", input_len_buffer)
        if len(serialized_tx)<input_len:
          raise Exception("Serialized transaction doesn't contain enough bytes for input %s"%_input_index)
        input_index_buffer, serialized_tx = serialized_tx[:input_len], serialized_tx[input_len:]

        if not skip_verification:
            if (not self.txos_storage.confirmed.has(input_index_buffer, rtx=rtx)):
              raise Exception("Unknown input index")
            self.inputs.append(self.txos_storage.confirmed.get(input_index_buffer, rtx=rtx))
        else:
            self.inputs.append(input_index_buffer)

    if len(serialized_tx)<2:
        raise Exception("Serialized transaction doesn't contain enough bytes for outputs array length")
    outputs_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
    (outputs_len,) = struct.unpack("> H", outputs_len_buffer)
    for _output_index in range(outputs_len):
      if len(serialized_tx)<2:
          raise Exception("Serialized transaction doesn't contain enough bytes for output %s length"%_output_index)
      output_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
      (output_len,) = struct.unpack("> H", output_len_buffer)
      if len(serialized_tx)<output_len:
          raise Exception("Serialized transaction doesn't contain enough bytes for output %s"%_output_index)
      output_buffer, serialized_tx = serialized_tx[:output_len], serialized_tx[output_len:]
      self.outputs.append(IOput(binary_object=output_buffer) )

    if len(serialized_tx)<2:
        raise Exception("Serialized transaction doesn't contain enough bytes for additional excesses array length")
    aes_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
    (aes_len,) = struct.unpack("> H", aes_len_buffer)
    for _ae in range(aes_len):
      if len(serialized_tx)<2:
          raise Exception("Serialized transaction doesn't contain enough bytes for additional excess %s length"%_ae)
      ae_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
      (ae_len,) = struct.unpack("> H", ae_len_buffer)
      if len(serialized_tx)<ae_len:
          raise Exception("Serialized transaction doesn't contain enough bytes for additional excess %s"%_ae)
      ae_buffer, serialized_tx = serialized_tx[:ae_len], serialized_tx[ae_len:]
      e = Excess.from_serialized(ae_buffer)
      self.additional_excesses.append(e )

    for _ue in range(inputs_len):
      if len(serialized_tx)<2:
          raise Exception("Serialized transaction doesn't contain enough bytes for updated excess %s length"%_ue)
      ue_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
      (ue_len,) = struct.unpack("> H", ue_len_buffer)
      if len(serialized_tx)<ue_len:
          raise Exception("Serialized transaction doesn't contain enough bytes for updated excess %s"%_ue)
      ue_buffer, serialized_tx = serialized_tx[:ue_len], serialized_tx[ue_len:]
      e = Excess.from_serialized(ue_buffer)
      #Depending on skif_verification self.inputs either contains inputs or serialized_indexes only
      if isinstance(self.inputs[_ue], bytes):
        self.updated_excesses[self.inputs[_ue]]=e        
      elif isinstance(self.inputs[_ue], IOput):
        self.updated_excesses[self.inputs[_ue].serialized_index]=e
    if len(serialized_tx)<32:
        raise Exception("Serialized transaction doesn't contain enough bytes for mixer_offset")
    self.mixer_offset, serialized_tx = int.from_bytes(serialized_tx[:32], "big"), serialized_tx[32:]
    if not skip_verification:
      self.verify(rtx=rtx)

  def to_json(self):
    pass #TODO

  def from_json(self):
    pass #TODO

  def non_context_verify(self, block_height):
    #Actually we partially use context via block_height. Consider renaming.
    try:
      if verification_cache[(self.serialize(), block_height)]:
        #We set coinbase during verification, thus if we scip verification
        #we need to set it manually. TODO (verification should be free from initialisation stuff)
        for output in self.outputs:
          if output.version==0:
            self.coinbase = output
        return verification_cache[(self.serialize(), block_height)]
    except KeyError:
      pass

    assert is_sorted(self.inputs, key= lambda _input: _input.authorized_pedersen_commitment.serialize()), "Inputs are not sorted"
    assert is_sorted(self.outputs, key= lambda _output: _output.authorized_pedersen_commitment.serialize()), "Outputs are not sorted"
    assert is_sorted(self.additional_excesses, key= lambda e: e.index), "Additional excesses are not sorted"

    assert len(self.inputs)==len(self.updated_excesses)
    for _input in self.inputs:
       assert _input.lock_height<block_height, "Timelocked input"
       s_i = _input.serialized_index
       assert s_i in self.updated_excesses, "Updated excesses do not contain update for address %s"%_input.address.to_text()
       assert _input.address.serialized_pubkey == self.updated_excesses[s_i].serialized_pubkey

    #Check that there are no duplicated outputs
    #TODO probably authorized????
    assert len(set([_output.unauthorized_pedersen_commitment.serialize() for _output in self.outputs]))==len(self.outputs), "Duplicated output"

    coinbase_num=0
    
    output_apcs = []

    for output in self.outputs:
        assert output.verify(), "Nonvalid output"
        _o_index =output.serialized_index
        output_apcs.append(_o_index[:33])
        if output.version==0:
            coinbase_num+=1
            self.coinbase = output
    assert coinbase_num<2, "More then one coinbase"

    for excess in self.additional_excesses:
        assert excess.verify(), "Nonvalid excess"
        #if not excess.message in output_apcs:
        #  return False
        #else:
        #  output_apcs.remove(excess.message)

    left_side, right_side = [], []

    _t = PublicKey()

    # Transaction should contain outputs (while it may not contain inputs)
    assert len(self.outputs), "Empty outputs"

    if len(self.inputs):
      _t.combine([_input.authorized_pedersen_commitment.to_public_key().public_key for _input in self.inputs])
      inputs_pedersen_commitment_sum = _t.to_pedersen_commitment()
      left_side.append(inputs_pedersen_commitment_sum)

    if len(self.outputs):
      _t.combine([_output.authorized_pedersen_commitment.to_public_key().public_key for _output in self.outputs])
      outputs_pedersen_commitment_sum = _t.to_pedersen_commitment()
      right_side.append(outputs_pedersen_commitment_sum)

      _t.combine([_output.address.pubkey.public_key for _output in self.outputs])
      outputs_excesses_sum = _t.to_pedersen_commitment()
      left_side.append(outputs_excesses_sum )

    if len(self.additional_excesses):
        _t.combine([excess.pubkey.public_key for excess in self.additional_excesses])
        additional_excesses_sum = _t.to_pedersen_commitment()
        left_side.append(additional_excesses_sum)

    if coinbase_num:
        minted_pc = PedersenCommitment(value_generator = default_generator)
        minted_pc.create(self.coinbase.value, b'\x00'*32)
        left_side.append(minted_pc)
     
     
    relay_fee = 0   
    for _output in self.outputs:
      if not _output.version==0:
        if _output.generator==default_generator_ser:
          relay_fee += _output.relay_fee

    new_outputs_fee = self.calc_new_outputs_fee()
    fee = relay_fee + new_outputs_fee

    negative_fee = False
    if fee<0:
      # It's okay, transaction has consumed so many inputs that it is profitable by itself
      # however we need to handle it manually: libsecp256k1 cannot handle negative value
      negative_fee = True
      fee = -fee

    if not fee==0:
      fee_pc = PedersenCommitment(value_generator = default_generator) #TODO think about fees for assets
      fee_pc.create(fee, b'\x00'*32)
      if negative_fee:
        left_side.append(fee_pc)
      else:
        right_side.append(fee_pc)

    mixer_pc = (Point(default_blinding_generator)*self.mixer_offset).to_pedersen_commitment() #TODO we should optimise here and generate fee_mixer pc
    right_side.append(mixer_pc)

    checker = PedersenCommitment()
    # For transaction which contains coinbase only, both sides will be empty
    if len(left_side) or len(right_side):
      sum_to_zero = checker.verify_sum(left_side, right_side)
      assert sum_to_zero, "Non-zero Pedersen commitments summ"

    if self.coinbase:
      info = self.coinbase.info()
      assert info['exp'] == -1, "Non-transparent coinbase"
      assert self.coinbase.lock_height >= block_height + coinbase_maturity,\
             "Wrong coinbase maturity timelock: %d should be %d"%(\
              self.coinbase.lock_height, block_height + coinbase_maturity)

    tx_skel = TransactionSkeleton(tx=self)
    assert len(tx_skel.serialize(rich_format=False))<50000, "Too big tx_skeleton"
    verification_cache[(self.serialize(), block_height)] = True
    return True

  def verify(self, rtx, block_height = None, block_version = 1, skip_non_context=False):
    """
     Transaction is valid if:
      0) inputs and outputs are sorted  (non context verification)
      1) all inputs are in utxo set
      2) no outputs are in utxo
      3) no updated_excesses duplicate already existed 
      3) all outputs are unique (exactly equal outputs are prohibited) (non context verification)
      3) all outputs are valid (non context verification)
      4) all additional excesses are valid (non context verification)
      5) sum(inputs)+sum(additional_excesses)+sum(outputs_excesses)+sum(minted_coins) == sum(outputs) + sum(fee) (non context verification)
      6) For each combined excess sum(digested_excesses) == combined_excess (non context verification)
      7) if coinbase is presented lock_height is set correctly (non context verification)
      8) size check : serialized tx_skeleton should have size less than 50k bytes (non context verification)
    """
    if not block_height:
      # Why +1 is here:
      # There are two general cases for verifying transactions:
      # 1) While generating new transaction. In this case transaction will be in the next block.
      # 2) While checking transaction in the block. In this case, current blockchain state is set 
      #     to prev block (prev to which we are checking) and block_height is current+1
      block_height = self.txos_storage.storage_space.blockchain.current_height(rtx=rtx) + 1

    if not skip_non_context: # skip_non_context is used when non context verification for that tx was made earlier
      assert self.non_context_verify(block_height)

    database_inputs = []
    assert len(set([_input.serialized_index for _input in self.inputs]))==len(self.inputs)
    for _input in self.inputs:
          index=_input.serialized_index
          if not self.txos_storage.confirmed.has(index, rtx=rtx):
              raise Exception("Spend unknown output")
          else:
              database_inputs.append(self.txos_storage.confirmed.get(index, rtx=rtx))
          assert not self.excesses_storage.excesses.has_index(rtx, self.updated_excesses[index].index), "Duplication of already existed excess during update"
          self.inputs = database_inputs
    
    for _output in self.outputs:
          _o_index = _output.serialized_index
          if self.txos_storage.confirmed.has(_o_index, rtx=rtx):
              raise Exception("Create duplicate output")
          elif _o_index in self.txos_storage.mempool:
              pass #It's ok
          else:
              #previously unknown output, let's add to database
              self.txos_storage.mempool[_o_index] = _output
          if not _output.block_version == block_version:
              raise Exception("Incompatible block version")

    for _ae in self.additional_excesses:
      assert not self.excesses_storage.excesses.has_index(rtx, _ae.index), "New additional excess duplicates old one"

    if block_height > 0:
      prev_block_props = {'height': self.txos_storage.storage_space.blockchain.current_height(rtx=rtx), 
                        'timestamp': self.txos_storage.storage_space.headers_storage.get(self.txos_storage.storage_space.blockchain.current_tip(rtx=rtx), rtx=rtx).timestamp}
    else:
      prev_block_props = {'height':0, 'timestamp':0}
    excess_lookup_partial = partial(excess_lookup, rtx=rtx, tx=self, excesses_storage = self.excesses_storage)
    assert evaluate_scripts(self, prev_block_props, excess_lookup_partial), "Bad script"
    assert check_burdens(self, self.txos_storage.confirmed.burden, self.excesses_storage, rtx=rtx), "Burden is not satisfied"

    return True
    #[inputs_pedersen_commitment_sum, outputs_excesses_sum, additional_excesses_sum], [outputs_pedersen_commitment_sum, fee_pc])
    #sum(inputs)+sum(additional_excesses)+sum(outputs_excesses) == sum(outputs) + sum(fee)
    

  def merge(self, another_tx, rtx):
    #self.serialized = None
    tx=Transaction(txos_storage = self.txos_storage, excesses_storage=self.excesses_storage)
    tx.inputs=self.inputs+another_tx.inputs
    tx.outputs=self.outputs+another_tx.outputs
    tx.additional_excesses = self.additional_excesses + another_tx.additional_excesses
    tx.updated_excesses = self.updated_excesses.copy()
    tx.updated_excesses.update(another_tx.updated_excesses)
    tx.mixer_offset = sum_offset(self.mixer_offset, another_tx.mixer_offset)
    tx.sort_lists()
    #TODO
    # If we merge transactions where the second spends outputs from the first, result is invalid
    # since we don't delete identical ioputs 
    assert tx.verify(rtx=rtx)
    return tx

  def __repr__(self):
    s=""
    s+="Inputs:\n"
    for _input in self.inputs:
       s+=" "*4+_input.__str__()+"\n"
    
    s+="Outputs:\n"
    for _output in self.outputs:
       s+=" "*4+_output.__str__()+"\n"
    return s

  @property
  def relay_fee(self):
    fee = 0
    for _output in self.outputs:
       fee += _output.relay_fee
    return fee

    



from leer.core.parameters.constants import bugged_outputs
from leer.core.storage.merkle_storage import MMR
from secp256k1_zkp import PedersenCommitment, PublicKey
from leer.core.lubbadubdub.address import Excess
import hashlib
import os
from leer.core.utils import sha256

class ExcessMMR(MMR):
      def sum(self, x1,x2):
        # each index is 33 bytes for commitments and 32 for hash
        pubkey1, hash1 = x1[:33], x1[33:65]
        pubkey2, hash2 = x2[:33], x2[33:65]
        pubkey1, pubkey2 = PublicKey(pubkey=pubkey1, raw=True), PublicKey(pubkey=pubkey2, raw=True)
        # consider pubkey1+pubkey2
        pk= PublicKey()
        pk.combine([pubkey1.public_key, pubkey2.public_key])
        first_part = pk.serialize()
        second_part = sha256(hash1+hash2)
        return first_part+second_part




class ExcessesStorage():
    '''
      Excesses tree:
        Each excess leaf obj is serialized excess, leaf index is excess_index.
    '''
    __shared_states = {}

    def __init__(self, storage_space, wtx):
      path = storage_space.path
      if not path in self.__shared_states:
        self.__shared_states[path]={}
      self.__dict__ = self.__shared_states[path]
      self.excesses = ExcessMMR("excesses", path, wtx=wtx, env=storage_space.env, clear_only=False)
      self.storage_space = storage_space
      storage_space.register_excesses_storage(self)
      
    def append_utxo(self, utxo, wtx):
      return self.excesses.append(wtx=wtx, obj_index=utxo.address.index, obj=b"")

    def append_additional_excess(self, excess, wtx):
      self.excesses.append_unique(wtx=wtx, obj_index=excess.index, obj=excess.serialize())

    def update_spent_address_with_excess(self, num_index, new_excess, wtx):
       new_excess_index = new_excess.index
       serialized_excess = new_excess.serialize()
       return self.update_spent_address_with_serialized_excess(num_index, new_excess_index, serialized_excess, wtx)

    def update_spent_address_with_serialized_excess(self, num_index_ser, new_excess_index, serialized_excess, wtx):
       old_excess_index, old_excess = self.excesses.update_index_by_num_unique(wtx, num_index_ser, new_excess_index, serialized_excess)
       if not new_excess_index[:33]==old_excess_index[:33]: #First 33 bytes - serialized pubkey
         raise Exception("Wrong excess update") #Never should get here, since tx is already checked, but this check is cheap
       return num_index_ser, old_excess_index, old_excess
      
    def rollback_serialized_excess_to_address(self, num_index_ser, address_index, serialized_address, wtx):
       excess_index, excess = self.excesses.update_index_by_num(wtx, num_index_ser, address_index, serialized_address)
       if not excess_index[:33]==address_index[:33]: #First 33 bytes - serialized pubkey
         raise Exception("Wrong excess update") #Never should get here, since tx is already checked, but this check is cheap
       return num_index_ser, excess_index, excess
      

    #def __contains__(self, serialized_index, ):
    #  return bool(self.excesses.get_by_hash(serialized_index))

    def has_index(self, serialized_index, rtx):
      return bool(self.excesses.get_by_hash(serialized_index, rtx=rtx))

    def get_by_index(self, serialized_index, rtx):
      serialized_excess = self.excesses.get_by_hash(serialized_index, rtx=rtx)
      if not serialized_excess:
        return None
      e = Excess()
      e.deserialize_raw(serialized_excess)
      return e

    def remove(self, n, wtx):
      self.excesses.remove(n, wtx=wtx)

    def get_root(self, rtx):
      return self.excesses.get_root(rtx=rtx)

    def apply_tx_get_merkles_and_rollback(self, tx, wtx):
      initial_state = self.get_state(rtx=wtx)
      initial_root = self.get_root(rtx=wtx)
      num_of_added_excesses, rollback_updates = self.apply_tx(tx, b"Excess validation temporal state", wtx)
      root = self.get_root(rtx=wtx)
      self.rollback(num_of_added_excesses, initial_state, rollback_updates, wtx)
      assert initial_root==self.get_root(rtx=wtx), "Database was corrupted during excesses apply_tx_get_merkles_and_rollback"
      return root

    #TODO bad naming. It should be apply block, or block_transaction
    def apply_tx(self, tx, new_state, wtx):
      for _o in tx.outputs:
          num = self.append_utxo(_o, wtx=wtx)
          _o.address_excess_num_index = num #storing num_index of stored address excess
      for _e in tx.additional_excesses:
          self.append_additional_excess(_e, wtx=wtx)
      rollback_updates = []
      for _i in tx.inputs:
          ind= _i.serialized_index
          rb = self.update_spent_address_with_excess(_i.address_excess_num_index, tx.updated_excesses[ind], wtx=wtx)
          rollback_updates.append(rb)
      self.set_state(new_state, wtx=wtx)
      return len(tx.additional_excesses)+len(tx.outputs), rollback_updates

    def rollback(self, num_of_added_excesses, prev_state, rollback_updates, wtx):
      self.excesses.remove(num_of_added_excesses, wtx=wtx)
      self.set_state(prev_state, wtx=wtx)
      for rb in rollback_updates:
        if rb[2]==None and rb[1] in bugged_outputs and rb[0] in bugged_outputs[rb[1]]:
          rb = (rb[0], rb[1], b"")
        self.rollback_serialized_excess_to_address(rb[0], rb[1], rb[2], wtx=wtx)

    def get_state(self, rtx):
      return self.excesses.get_state(rtx=rtx)

    def set_state(self, state, wtx):
      self.excesses.set_state(state, wtx=wtx)
